/* gcomposer: a little tool to write music.
 * http://sed.free.fr/gcomposer
 * Released in the public domain (freedom has no license).
 */

#include "doc.h"

#include <stdarg.h>
#include <string.h>
#include <stdlib.h>

#include "draw.h"
#include "mem.h"
#include "error.h"
#include "dim.h"
#include "dprintf.h"

/* return new_pos + 1 (for 0 to be interpreted as special case) */
static int put_note_in_place(chord *c, int note)
{
  int pos;
  int note_pos;
  elem *e;
  elem dum;

  e = &c->u.chord.data[note];
  switch (e->type) {
    case NOTE: note_pos = e->u.note.pitch; break;
    case SILENCE: note_pos = e->u.silence.position; break;
  }

  /* go closer to the head? */
  while (note) {
    e = &c->u.chord.data[note-1];
    switch (e->type) {
      case NOTE: pos = e->u.note.pitch; break;
      case SILENCE: pos = e->u.silence.position; break;
    }
    if (pos <= note_pos) break;
    dum = c->u.chord.data[note-1];
    c->u.chord.data[note-1] = c->u.chord.data[note];
    c->u.chord.data[note] = dum;
    note--;
  }

  /* go closer to the tail? */
  while (note < c->u.chord.size-1) {
    e = &c->u.chord.data[note+1];
    switch (e->type) {
      case NOTE: pos = e->u.note.pitch; break;
      case SILENCE: pos = e->u.silence.position; break;
    }
    if (pos >= note_pos) break;
    dum = c->u.chord.data[note+1];
    c->u.chord.data[note+1] = c->u.chord.data[note];
    c->u.chord.data[note] = dum;
    note++;
  }

  return note + 1;
}

int doc_insert_note(doc *d, int line, int ch, elem_type type, ...)
{
  chord *c = &d->data[line].u.music.data[ch];
  va_list va;
  int n;

  if (!alloc_note(d, line, ch)) return 0;

  NOTE_TYPE(d, line, ch, c->u.chord.size-1) = type;

  va_start(va, type);

  n = c->u.chord.size-1;
  switch(type) {
    case NOTE:
      NOTE_PITCH(d, line, ch, n) = va_arg(va, int);
      NOTE_ALTERATION(d, line, ch, n) = va_arg(va, note_alteration);
      NOTE_DURATION(d, line, ch, n) = va_arg(va, note_duration);
      NOTE_NB_DOTS(d, line, ch, n) = va_arg(va, int);
      NOTE_HEAD_DELTA(d, line, ch, n) = va_arg(va, int);
      NOTE_ALTERATION_DELTA(d, line, ch, n) = va_arg(va, int);
      NOTE_DOTS_DELTA(d, line, ch, n) = va_arg(va, int);
      NOTE_STEM_DOWN(d, line, ch, n) = va_arg(va, int);
      break;
    case SILENCE:
      SILENCE_POSITION(d, line, ch, n) = va_arg(va, int);
      SILENCE_DURATION(d, line, ch, n) = va_arg(va, note_duration);
      SILENCE_NB_DOTS(d, line, ch, n) = va_arg(va, int);
      break;
  }

  va_end(va);

  /*invalidate_line(d, line);*/

  return put_note_in_place(c, c->u.chord.size-1);
}

int doc_set_note_pitch(doc *d, int line, int chord, int note, int pitch)
{
  if (pitch > STAFF_NOTE_MAX) return 0;
  if (pitch < STAFF_NOTE_MIN) return 0;
  NOTE_PITCH(d, line, chord, note) = pitch;
  return put_note_in_place(&CHORD(d, line, chord), note);
}

int doc_set_note_position(doc *d, int line, int chord, int note, int position)
{
  if (position > STAFF_NOTE_MAX) return 0;
  if (position < STAFF_NOTE_MIN) return 0;
  SILENCE_POSITION(d, line, chord, note) = position;
  return put_note_in_place(&CHORD(d, line, chord), note);
}

int doc_set_note_alteration(doc *d, int line, int chord, int note,
                            note_alteration t)
{
  NOTE_ALTERATION(d, line, chord, note) = t;
  return 1;
}

int doc_set_note_duration(doc *d, int line, int chord, int note,
                          note_duration t)
{
  switch (NOTE_TYPE(d, line, chord, note)) {
    case NOTE:
      if (t == NOTE_DURATION(d, line, chord, note)) return 0;
      NOTE_DURATION(d, line, chord, note) = t;
      break;
    case SILENCE:
      if (t == SILENCE_DURATION(d, line, chord, note)) return 0;
      SILENCE_DURATION(d, line, chord, note) = t;
      break;
  }
  return 1;
}

int doc_set_note_dot(doc *d, int line, int chord, int note, int nb_dots)
{
  switch (NOTE_TYPE(d, line, chord, note)) {
    case NOTE:
      NOTE_NB_DOTS(d, line, chord, note) = nb_dots;
      break;
    case SILENCE:
      SILENCE_NB_DOTS(d, line, chord, note) = nb_dots;
      break;
  }
  return 1;
}

int doc_set_note_head_delta(doc *d, int line, int chord, int note, int delta)
{
  if (delta < MIN_NOTE_HEAD_DELTA || delta > MAX_NOTE_HEAD_DELTA) return 0;
  NOTE_HEAD_DELTA(d, line, chord, note) = delta;
  return 1;
}

int doc_set_note_alteration_delta(doc *d, int line, int chord, int note,
                                  int delta)
{
  if (delta < MIN_ALTERATION_DELTA || delta > MAX_ALTERATION_DELTA) return 0;
  NOTE_ALTERATION_DELTA(d, line, chord, note) = delta;
  return 1;
}

int doc_set_note_dots_delta(doc *d, int line, int chord, int note, int delta)
{
  if (delta < MIN_DOTS_DELTA || delta > MAX_DOTS_DELTA) return 0;
  NOTE_DOTS_DELTA(d, line, chord, note) = delta;
  return 1;
}

int doc_set_note_stem_down(doc *d, int line, int chord, int note, int down)
{
  /* !! here because stem_down is a int:1 which has value -1, not 1 */
  if (down == !!NOTE_STEM_DOWN(d, line, chord, note)) return 0;
  NOTE_STEM_DOWN(d, line, chord, note) = down;
  return 1;
}

int doc_set_tonalite_position(doc *d, int line, int chord, int position)
{
  if (position > STAFF_NOTE_MAX) return 0;
  if (position < STAFF_NOTE_MIN) return 0;
  TONALITE_POSITION(d, line, chord) = position;
  return 1;
}

int doc_set_tonalite_alteration(doc *d, int line, int chord,
                                note_alteration a)
{
  TONALITE_ALTERATION(d, line, chord) = a;
  return 1;
}

int doc_set_clef_type(doc *d, int line, int chord, clef_type a)
{
  CLEF_TYPE(d, line, chord) = a;
  return 1;
}

int doc_set_clef_position(doc *d, int line, int chord, int position)
{
  if (position > CLEF_POS_MAX) return 0;
  if (position < CLEF_POS_MIN) return 0;
  CLEF_POSITION(d, line, chord) = position;
  return 1;
}

/* before_chord == -1 to put at the end */
int doc_insert_chord(doc *d, int line_number, int before_chord,
                     chord_type type)
{
  line *l;
  chord c;
  int i;

  /* create new chord */
  if (!alloc_chord(d, line_number)) return 0;

  l = &d->data[line_number];

  CHORD(d, line_number, l->u.music.size-1).type = type;

  if (type == TONALITE) {
    TONALITE_POSITION(d, line_number, l->u.music.size-1) = 0;
    TONALITE_ALTERATION(d, line_number, l->u.music.size-1) = DIESE;
  } else if (type == CLEF) {
    CLEF_TYPE(d, line_number, l->u.music.size-1) = CLEF_G;
    CLEF_POSITION(d, line_number, l->u.music.size-1) = 0;
  }

  /* insert new chord */
  if (before_chord != -1) {
    /* we move chords, not very efficient but a line does not have
     * thousands of chords, does it?
     */
    c = l->u.music.data[l->u.music.size-1];
    for (i=l->u.music.size-2; i>=before_chord; i--)
      l->u.music.data[i+1] = l->u.music.data[i];
    l->u.music.data[before_chord] = c;
  }

  return 1;
}

/* give after_line == -1 to insert at the head */
int doc_insert_line(doc *d, line_type type, int after_line)
{
  line l;

  /* create new line */
  if (!alloc_line(d)) return 0;

  /* insert new line */
  if (d->size-1) {
    l = d->data[d->size-1];
    memmove(d->data+after_line+2, d->data+after_line+1,
            (d->size-2-after_line) * sizeof(line));
    d->data[after_line+1] = l;
  }

  LINE_TYPE(d, after_line+1) = type;

  return 1;
}

int doc_merge_lines(doc *d, int first_line, merge_enum del)
{
  line *l = d->data+first_line;
  line *l2 = d->data+first_line+1;
  int i;
  int nb_chords = l2->u.music.size;
  int saved_size = l->u.music.size;

  for (i = 0; i < nb_chords; i++) {
    if (!alloc_chord(d, first_line)) goto copy_error;
    memcpy(l->u.music.data+l->u.music.size-1, l2->u.music.data+i,
           sizeof(chord));
  }

  l2->u.music.size = 0;

  switch (del) {
    case MERGE_DELETE:
      /* do nothing */
      break;
    case MERGE_DONT_DELETE:
      /* remove data so no free is done */
      l2->u.music.maxsize = 0;
      l2->u.music.data = NULL;
      break;
  }

  doc_delete_line(d, first_line+1);

  return 1;

copy_error:
  l->u.music.size = saved_size;
  return 0;
}

/* you can pass a buffer to use, * pass NULL if buffer has to be allocated */
int doc_split_line(doc *d, int line_number, int before_chord,
                   void *buf, int buf_size)
{
  line *l;
  line *l2;
  int i;
  int nb_chords;

  if (!doc_insert_line(d, LINE_MUSIC, line_number)) return 0;

  l = d->data+line_number;
  nb_chords = l->u.music.size - before_chord;

  l2 = d->data+line_number+1;

  /* put buffer if there is one */
  if (buf) {
    l2->u.music.data = buf;
    l2->u.music.maxsize = buf_size;
  }

  for (i = 0; i < nb_chords; i++) {
    if (!alloc_chord(d, line_number+1)) goto copy_error;
    memcpy(l2->u.music.data+i, l->u.music.data+i+before_chord, sizeof(chord));
  }

  l->u.music.size -= nb_chords;

  return 1;

copy_error:
  l2->u.music.size = 0;
  doc_delete_line(d, line_number+1);
  return 0;
}

int doc_delete_line(doc *d, int line_number)
{
  int i;
  line *l = d->data+line_number;

  switch (LINE_TYPE(d, line_number)) {
    case LINE_MUSIC:
      for (i = 0; i < l->u.music.size; i++)
        doc_delete_chord(d, line_number, i);

      if (l->u.music.maxsize) {
        free(l->u.music.data);
        l->u.music.maxsize = 0;
        l->u.music.size = 0;
        l->u.music.data = NULL;
      }
      break;
    case LINE_TEXT:
      if (LINE_TEXT_DATA(d, line_number)) {
        free(LINE_TEXT_DATA(d, line_number));
        LINE_TEXT_DATA(d, line_number) = NULL;
      }
  }

  /* this is too slow */
  memmove(d->data+line_number, d->data+line_number+1,
          sizeof(line) * (d->size - line_number-1));

  d->size--;

  return 1;
}

int doc_delete_chord(doc *d, int line_number, int chord_number)
{
  line *l = d->data+line_number;
  chord *c = l->u.music.data+chord_number;

  if (c->type == NOTES_CHORD && c->u.chord.maxsize) {
    free(c->u.chord.data);
    c->u.chord.maxsize = 0;
    c->u.chord.size = 0;
    c->u.chord.data = NULL;
  }

  /* too slow */
  memmove(l->u.music.data+chord_number, l->u.music.data+chord_number+1,
          sizeof(chord) * (l->u.music.size-chord_number-1));

  l->u.music.size--;

  return 1;
}

int doc_delete_note(doc *d, int line_number, int chord_number, int note)
{
  line *l = d->data+line_number;
  chord *c = l->u.music.data+chord_number;

  /* too slow */
  memmove(c->u.chord.data+note, c->u.chord.data+note+1,
          sizeof(elem) * (c->u.chord.size-note-1));

  c->u.chord.size--;

  return 1;
}

/* how ugly, O(n) to destroy stuff, shame on me */
void destroy_doc(doc *d)
{
  int i;

  for (i = 0; i < d->size; i++) {
    line *l = d->data + i;
    int j;

    switch (l->type) {
      case LINE_MUSIC:
        for (j = 0; j < l->u.music.size; j++) {
          chord *c = l->u.music.data + j;
          if (c->type == NOTES_CHORD) {
            if (c->u.chord.maxsize) {
              dprintf("freeing %p\n", c->u.chord.data);
              free (c->u.chord.data);
            }

            c->u.chord.size = 0;
            c->u.chord.maxsize = 0;
            c->u.chord.data = NULL;
          }
        }

        if (l->u.music.maxsize) {
          dprintf("freeing %p\n", l->u.music.data);
          free(l->u.music.data);
        }

        l->u.music.size = 0;
        l->u.music.maxsize = 0;
        l->u.music.data = NULL;
        break;
      case LINE_TEXT:
        if (l->u.text.data) {
          dprintf("freeing %p\n", l->u.text.data);
          free(l->u.text.data);
          l->u.text.data = NULL;
        }
    }
  }

  if (d->maxsize) {
    dprintf("freeing %p\n", d->data);
    free(d->data);
  }

  d->size = 0;
  d->maxsize = 0;
  d->data = NULL;
  d->filename[0] = 0;

  destroy_history(d);
}

void destroy_line(line *l)
{
  int i;

  switch (l->type) {
    case LINE_MUSIC:
      for (i = 0; i < l->u.music.size; i++) {
        chord *c = l->u.music.data + i;
        if (c->type == NOTES_CHORD) {
          if (c->u.chord.maxsize) free (c->u.chord.data);

          c->u.chord.size = 0;
          c->u.chord.maxsize = 0;
          c->u.chord.data = NULL;
        }
      }

      if (l->u.music.maxsize) free(l->u.music.data);

      l->u.music.size = 0;
      l->u.music.maxsize = 0;
      l->u.music.data = NULL;
      break;
    case LINE_TEXT:
      if (l->u.text.data) free(l->u.text.data);
      l->u.text.data = NULL;
      break;
  }
}

void destroy_chord(chord *c)
{
  switch (c->type) {
    case NOTES_CHORD:
      if (c->u.chord.maxsize == 0) return;
      free(c->u.chord.data);
      c->u.chord.size = 0;
      c->u.chord.maxsize = 0;
      c->u.chord.data = NULL;
      return;
    case MESURE_BREAK: return;
    case TONALITE: return;
    case CLEF: return;
  }
}

int copy_chord(chord *to, chord *from)
{
  to->type = from->type;

  switch (to->type) {
    case NOTES_CHORD:
      to->u.chord.maxsize = 0;
      to->u.chord.size = 0;
      to->u.chord.data = NULL;
      if (from->u.chord.size == 0) return 1;
      to->u.chord.data = malloc(from->u.chord.size * sizeof(elem));
      if (!to->u.chord.data) {
        set_error(OOM);
        return 0;
      }
      to->u.chord.size = from->u.chord.size;
      to->u.chord.maxsize = from->u.chord.size;
      memcpy(to->u.chord.data, from->u.chord.data,
             to->u.chord.size * sizeof(elem));
      return 1;
    case MESURE_BREAK:
      return 1;
    case TONALITE:
      to->u.tonalite.position = from->u.tonalite.position;
      to->u.tonalite.alteration = from->u.tonalite.alteration;
      return 1;
    case CLEF:
      to->u.clef.type = from->u.clef.type;
      to->u.clef.position = from->u.clef.position;
      return 1;
  }

  /* dummy return for no gcc warning */
  return 1;
}

int get_delta_clef(doc *d, int l, int c)
{
  int delta_clef = 0;
  int i;

  /* see play.c:start_chord for the explaination of the values */
  for (i = 0; i < c; i++) {
    if (CHORD_TYPE(d, l, i) != CLEF) continue;
    switch (CLEF_TYPE(d, l, i)) {
      case CLEF_C: delta_clef = 4; break;
      case CLEF_F: delta_clef = 8; break;
      case CLEF_G: delta_clef = 0; break;
      case CLEF_G_8va: delta_clef = 7; break;
      case CLEF_G_8vb: delta_clef = -7; break;
    }
    delta_clef += CLEF_POSITION(d, l, i);
  }

  return delta_clef;
}
