/* gcomposer: a little tool to write music.
 * http://sed.free.fr/gcomposer
 * Released in the public domain (freedom has no license).
 */

#include "draw.h"

#include "gtk_wrapper.h"
#include "images.h"
#include "cursor.h"
#include "size.h"
#include "dim.h"
#include "dprintf.h"

static void draw_up_lines(void *drawing_area, int x, int y, int nb, int zoom)
{
  int i;
  for (i = 0; i < nb; i++)
    gtk_wrapper_draw_line(drawing_area,
                          x - MINILINE_WIDTH/2, y - (i + 1) * LINE_SPACING,
                          x + MINILINE_WIDTH/2, y - (i + 1) * LINE_SPACING,
                          zoom);
}

static void draw_down_lines(void *drawing_area, int x, int y, int nb, int zoom)
{
  int i;
  for (i = 0; i < nb; i++)
    gtk_wrapper_draw_line(drawing_area,
                          x - MINILINE_WIDTH / 2,
                          y + i * LINE_SPACING + 5 * LINE_SPACING,
                          x + MINILINE_WIDTH / 2,
                          y + i * LINE_SPACING + 5 * LINE_SPACING,
                          zoom);
}

static void draw_silence(void *drawing_area, elem *e, int x, int y, int zoom)
{
  int note_center_x;
  int note_center_y;
  enum image_type t;
  int i;
  int adjust = 0;

  note_center_y = y + 3 * LINE_SPACING
                  - e->u.silence.position * LINE_SPACING/2;
  note_center_x = x;

  /* draw the silence */
  switch (e->u.silence.duration) {
    case RONDE: t = I_SILENCE_RONDE; adjust=RONDE_SILENCE_ADJUST; break;
    case BLANCHE: t = I_SILENCE_BLANCHE; adjust=BLANCHE_SILENCE_ADJUST; break;
    case NOIRE: t = I_SILENCE_NOIRE; break;
    case CROCHE: t = I_SILENCE_CROCHE; break;
    case DCROCHE: t = I_SILENCE_DCROCHE; break;
    case DDCROCHE: t = I_SILENCE_DDCROCHE; break;
    case DDDCROCHE: t = I_SILENCE_DDDCROCHE; break;
    case DDDDCROCHE: t = I_SILENCE_DDDDCROCHE; break;
  }

  gtk_wrapper_draw_image(drawing_area, t,
                         note_center_x + e->dim.min_x,
                         note_center_y + e->dim.min_y,
                         zoom);

  /* draw dots */
  for (i = 0; i < e->u.silence.nb_dots; i++) {
    gtk_wrapper_draw_image(drawing_area, I_DOT,
                           note_center_x + images[t].width / 2
                                     + i * (DOT_SPACE + images[I_DOT].width)
                                     + DOT_SILENCE_SPACE,
                           note_center_y - images[I_DOT].height / 2 + adjust,
                           zoom);
  }
}

static void draw_note(void *drawing_area, elem *e, int x, int y, int zoom)
{
  int note_center_x;
  int note_center_y;
  int i;

  note_center_y = y + 3 * LINE_SPACING - e->u.note.pitch * LINE_SPACING/2;
  note_center_x = x;

  /* head delta stuff */
  note_center_x += e->u.note.head_delta;

  /* draw head */
  switch (e->u.note.duration) {
    case RONDE:
    case BLANCHE:
      gtk_wrapper_draw_image(drawing_area, I_WHITE_HEAD,
                             note_center_x - images[I_WHITE_HEAD].width / 2,
                             note_center_y - images[I_WHITE_HEAD].height / 2,
                             zoom);
      break;
    default:
      gtk_wrapper_draw_image(drawing_area, I_BLACK_HEAD,
                             note_center_x - images[I_BLACK_HEAD].width / 2,
                             note_center_y - images[I_BLACK_HEAD].height / 2,
                             zoom);
      break;
  }

  /* draw diese/bemol/becarre */
  /* this code is not correct, I_BLACK_HEAD should be used only
   * with corresponding head (anyway normally white and black
   * heads have save width/height)
   */
  switch (e->u.note.alteration) {
    case NORMAL: break;
    case DIESE:
      gtk_wrapper_draw_image(drawing_area, I_DIESE,
                             note_center_x - images[I_BLACK_HEAD].width / 2
                                     - images[I_DIESE].width - DIESE_SPACE
                                     + e->u.note.alteration_delta,
                             note_center_y - images[I_DIESE].height / 2,
                             zoom);
      break;
    case BEMOL:
      gtk_wrapper_draw_image(drawing_area, I_BEMOL,
                             note_center_x - images[I_BLACK_HEAD].width / 2
                                     - images[I_BEMOL].width - BEMOL_SPACE
                                     + e->u.note.alteration_delta,
                             note_center_y - images[I_BEMOL].height / 2
                                           - BEMOL_ADJUST,
                             zoom);
      break;
    case BECARRE:
      gtk_wrapper_draw_image(drawing_area, I_BECARRE,
                             note_center_x - images[I_BLACK_HEAD].width / 2
                                     - images[I_BECARRE].width - BEMOL_SPACE
                                     + e->u.note.alteration_delta,
                             note_center_y - images[I_BECARRE].height / 2,
                             zoom);
      break;
  }

  /* draw dots */
  for (i = 0; i < e->u.note.nb_dots; i++) {
    gtk_wrapper_draw_image(drawing_area, I_DOT,
                           note_center_x + images[I_BLACK_HEAD].width / 2
                                     + i * (DOT_SPACE + images[I_DOT].width)
                                     + DOT_NOTE_SPACE
                                     + e->u.note.dots_delta,
                           note_center_y - images[I_DOT].height / 2,
                           zoom);
  }
}

static void draw_elem(void *drawing_area, elem *e, int x, int y, int zoom)
{
  switch (e->type) {
    case SILENCE: return draw_silence(drawing_area, e, x, y, zoom);
    case NOTE: return draw_note(drawing_area, e, x, y, zoom);
  }
}

static void draw_chord(void *drawing_area, chord *c, int x, int y, int zoom)
{
  int i;

  x -= c->dim.min_x;

  switch (c->type) {
    case NOTES_CHORD: break;
    case MESURE_BREAK:
      gtk_wrapper_draw_line(drawing_area, x, y, x, y+4*LINE_SPACING-1, zoom);
      gtk_wrapper_draw_line(drawing_area, x+1, y, x+1, y+4*LINE_SPACING-1,
                            zoom);
      return;
    case TONALITE: {
      int type;
      /* draw supplementary lines if needed */
      if (c->u.tonalite.position >= 8)
        draw_up_lines(drawing_area, x, y, (c->u.tonalite.position - 6) / 2,
                      zoom);
      else if (c->u.tonalite.position <= -4)
        draw_down_lines(drawing_area, x, y, (-c->u.tonalite.position - 2) / 2,
                        zoom);
      switch (c->u.tonalite.alteration) {
        case NORMAL: /* we can't come here */ return;
        case DIESE: type = I_DIESE; break;
        case BEMOL: type = I_BEMOL; break;
        case BECARRE: type = I_BECARRE; break;
      }
      gtk_wrapper_draw_image(drawing_area, type,
                             x - images[type].width/2, y + c->dim.min_y,
                             zoom);
      return;
    }
    case CLEF: {
      int type;
      switch (c->u.clef.type) {
        case CLEF_C: type = I_CLEF_C; break;
        case CLEF_F: type = I_CLEF_F; break;
        case CLEF_G: type = I_CLEF_G; break;
        case CLEF_G_8va: type = I_CLEF_G_8va; break;
        case CLEF_G_8vb: type = I_CLEF_G_8vb; break;
      }
      gtk_wrapper_draw_image(drawing_area, type,
                             x + c->dim.min_x, y + c->dim.min_y,
                             zoom);
      return;
    }
  }

  /* draw supplementary lines if needed */
  if (c->u.chord.size) {
    int position;
    elem *e = &c->u.chord.data[0];
    /* down lines? */
    switch (e->type) {
      case NOTE: position = e->u.note.pitch; break;
      case SILENCE: position = e->u.silence.position; break;
    }
    if (position <= -4)
      draw_down_lines(drawing_area, x, y, (-position - 2) / 2,
                      zoom);
    /* up lines? */
    e = &c->u.chord.data[c->u.chord.size-1];
    switch (e->type) {
      case NOTE: position = e->u.note.pitch; break;
      case SILENCE: position = e->u.silence.position; break;
    }
    if (position >= 8)
      draw_up_lines(drawing_area, x, y, (position - 6) / 2,
                    zoom);
  }

  /* draw every elements */
  for (i = 0; i < c->u.chord.size; i++)
    draw_elem(drawing_area, &c->u.chord.data[i], x, y, zoom);

  /* draw up stem */
  switch (c->dim.stem_up_type) {
    case RONDE:
    case BLANCHE:
    case NOIRE:
      break;
    case CROCHE:
      gtk_wrapper_draw_image(drawing_area, I_UP_CROCHE,
                             x+images[I_BLACK_HEAD].width/2,
                             y + c->dim.stem_up_min_y,
                             zoom);
      break;
    case DCROCHE:
      gtk_wrapper_draw_image(drawing_area, I_UP_DCROCHE,
                             x+images[I_BLACK_HEAD].width/2,
                             y + c->dim.stem_up_min_y,
                             zoom);
      break;
    case DDCROCHE:
      gtk_wrapper_draw_image(drawing_area, I_UP_DDCROCHE,
                             x+images[I_BLACK_HEAD].width/2,
                             y + c->dim.stem_up_min_y,
                             zoom);
      break;
    case DDDCROCHE:
      gtk_wrapper_draw_image(drawing_area, I_UP_DDDCROCHE,
                             x+images[I_BLACK_HEAD].width/2,
                             y + c->dim.stem_up_min_y,
                             zoom);
      break;
    case DDDDCROCHE:
      gtk_wrapper_draw_image(drawing_area, I_UP_DDDDCROCHE,
                             x+images[I_BLACK_HEAD].width/2,
                             y + c->dim.stem_up_min_y,
                             zoom);
      break;
  }
  if (c->dim.stem_up_type == RONDE) goto after_stem_up;
  else if (c->dim.stem_up_type == BLANCHE)
    gtk_wrapper_draw_line(drawing_area,
                          x+images[I_WHITE_HEAD].width/2,
                          y + c->dim.stem_up_min_y,
                          x+images[I_WHITE_HEAD].width/2,
                          y + c->dim.stem_up_max_y,
                          zoom);
  else
    gtk_wrapper_draw_line(drawing_area,
                          x+images[I_BLACK_HEAD].width/2,
                          y + c->dim.stem_up_min_y,
                          x+images[I_BLACK_HEAD].width/2,
                          y + c->dim.stem_up_max_y,
                          zoom);

after_stem_up:

  /* draw down stem */
  switch (c->dim.stem_down_type) {
    case RONDE:
    case BLANCHE:
    case NOIRE:
      break;
    case CROCHE:
      gtk_wrapper_draw_image(drawing_area, I_DOWN_CROCHE,
                             x-images[I_BLACK_HEAD].width/2,
                             y + c->dim.stem_down_max_y
                               - (images[I_DOWN_CROCHE].height-1),
                             zoom);
      break;
    case DCROCHE:
      gtk_wrapper_draw_image(drawing_area, I_DOWN_DCROCHE,
                             x-images[I_BLACK_HEAD].width/2,
                             y + c->dim.stem_down_max_y
                               - (images[I_DOWN_DCROCHE].height-1),
                             zoom);
      break;
    case DDCROCHE:
      gtk_wrapper_draw_image(drawing_area, I_DOWN_DDCROCHE,
                             x-images[I_BLACK_HEAD].width/2,
                             y + c->dim.stem_down_max_y
                               - (images[I_DOWN_DDCROCHE].height-1),
                             zoom);
      break;
    case DDDCROCHE:
      gtk_wrapper_draw_image(drawing_area, I_DOWN_DDDCROCHE,
                             x-images[I_BLACK_HEAD].width/2,
                             y + c->dim.stem_down_max_y
                               - (images[I_DOWN_DDDCROCHE].height-1),
                             zoom);
      break;
    case DDDDCROCHE:
      gtk_wrapper_draw_image(drawing_area, I_DOWN_DDDDCROCHE,
                             x-images[I_BLACK_HEAD].width/2,
                             y + c->dim.stem_down_max_y
                               - (images[I_DOWN_DDDDCROCHE].height-1),
                             zoom);
      break;
  }
  if (c->dim.stem_down_type == RONDE) goto after_stem_down;
  else if (c->dim.stem_down_type == BLANCHE)
    gtk_wrapper_draw_line(drawing_area,
                          x-images[I_WHITE_HEAD].width/2,
                          y + c->dim.stem_down_min_y,
                          x-images[I_WHITE_HEAD].width/2,
                          y + c->dim.stem_down_max_y,
                          zoom);
  else
    gtk_wrapper_draw_line(drawing_area,
                          x-images[I_BLACK_HEAD].width/2,
                          y + c->dim.stem_down_min_y,
                          x-images[I_BLACK_HEAD].width/2,
                          y + c->dim.stem_down_max_y,
                          zoom);

after_stem_down:;
}

void draw_line(doc *d, int i, int start_x, int end_x)
{
  line *l = &d->data[i];
  int x;
  int y;
  int width;
  int k;
  int first_chord;
  int last_chord;

  x = LEFT_MARGIN;
  y = l->dim.y - l->dim.min_y;

  /* get the width of the line */
  width = l->dim.width;

  switch (l->type) {
    case LINE_MUSIC:
      /* get chords */
      get_crossed_chords(l, start_x, end_x, &first_chord, &last_chord);
      dprintf("crossed_chords first %d last %d startx %d endx %d\n",
              first_chord, last_chord, start_x, end_x);

      /* draw the 5 lines */
      gtk_wrapper_draw_line(d->drawing_area, x, y, x+width-1, y, d->zoom);
      gtk_wrapper_draw_line(d->drawing_area, x, y+LINE_SPACING,
                            x+width-1, y+LINE_SPACING, d->zoom);
      gtk_wrapper_draw_line(d->drawing_area, x, y+LINE_SPACING*2,
                            x+width-1, y+LINE_SPACING*2, d->zoom);
      gtk_wrapper_draw_line(d->drawing_area, x, y+LINE_SPACING*3,
                            x+width-1, y+LINE_SPACING*3, d->zoom);
      gtk_wrapper_draw_line(d->drawing_area, x, y+LINE_SPACING*4,
                            x+width-1, y+LINE_SPACING*4, d->zoom);

      if (first_chord == -1) goto after_chord;

      /* draw the chords */
      for (k = first_chord; k <= last_chord; k++) {
        /* set the color */
        if (in_play(d, i, k))
          gtk_set_color(COLOR_GREEN);
        else if (in_selection(d, i, k))
          gtk_set_color(COLOR_BLUE);
        draw_chord(d->drawing_area, &l->u.music.data[k],
                   l->u.music.data[k].dim.x + LEFT_MARGIN, y, d->zoom);
        gtk_set_color(COLOR_BLACK);
      }

after_chord:
      break;
    case LINE_TEXT:
      if (in_selection(d, i, 0))
        gtk_set_color(COLOR_BLUE);
      gtk_wrapper_draw_string(d->drawing_area, d->fonts,
                              l->u.text.data,
                              x + TEXT_LEFT_BORDER,
                              y + TEXT_UPPER_MARGIN,
                              d->zoom);
      gtk_set_color(COLOR_BLACK);
      break;
  }

  /* draw the cursor if it's on the line */
  if (d->cursor.line == i) draw_cursor(&d->cursor);
}

void draw_area(doc *d, int x, int y, int w, int h)
{
  int start_line;
  int end_line;
  int i;

  get_crossed_lines(d, y, y+h-1, &start_line, &end_line);
  dprintf("draw start l %d end l %d (x=%d y=%d w=%d h=%d\n",
          start_line, end_line, x, y, w, h);
  if (start_line == -1) return;

  for (i = start_line; i <= end_line; i++)
    draw_line(d, i, x, x+w-1);

  /* don't allow shrink */
  update_horizontal_scrollbar(d);
}

void invalidate_line(doc *d, int i)
{
  line *l;
  int w;
  int h;

  dprintf("invalidate_line %d\n", i);

  ensure_line_positioned(d, i);

  l = &d->data[i];

  w = gtk_wrapper_get_width(d->drawing_area, d->zoom);
  h = l->dim.height;

  gtk_wrapper_widget_queue_draw_area(d->drawing_area,
      0, l->dim.y, w, h, d->zoom);
}

void invalidate_line_and_next(doc *d, int i)
{
  line *l;
  int w;
  int h;

  dprintf("invalidate_line_and_next %d\n", i);
  ensure_line_positioned(d, i);

  l = &d->data[i];

  w = gtk_wrapper_get_width(d->drawing_area, d->zoom);
  /* we take the full height, it's more than the size of the
   * drawing if l is not the first line, but hey, it works
   */
  h = gtk_wrapper_get_height(d->drawing_area, d->zoom);

  gtk_wrapper_widget_queue_draw_area(d->drawing_area,
      0, l->dim.y, w, h, d->zoom);
}

/* invalidate the area after the last line
 * (used when deleting lines)
 */
void invalidate_next(doc *d)
{
  line *l;
  int w;
  int h;

  dprintf("invalidate_next\n");
  ensure_line_positioned(d, d->size-1);

  l = &d->data[d->size-1];

  w = gtk_wrapper_get_width(d->drawing_area, d->zoom);
  /* we take the full height, it's more than the size of the
   * drawing if l is not the first line, but hey, it works
   */
  h = gtk_wrapper_get_height(d->drawing_area, d->zoom);

  gtk_wrapper_widget_queue_draw_area(d->drawing_area,
      0, l->dim.y + l->dim.height, w, h, d->zoom);
}

void invalidate_all(doc *d)
{
  gtk_wrapper_widget_queue_draw(d->drawing_area);
}

static void clear_cursor_on_music_line(cursor *c)
{
  int x;
  int y;
  int height;
  line *l;
  int pos;
  int has_to_draw;

  dprintf("clear_cursor\n");

  //if (c->hidden_level) return;
  //if (!c->state) return;

  ensure_line_positioned(c->doc, c->line);

  l = ((doc *)c->doc)->data + c->line;
  if (c->chord < l->u.music.size)
    x = LEFT_MARGIN +
        l->u.music.data[c->chord].dim.x - 3;
  else if (l->u.music.size)
    x = LEFT_MARGIN +
        l->u.music.data[l->u.music.size-1].dim.x +
        l->u.music.data[l->u.music.size-1].dim.width - 3;
  else
    x = LEFT_MARGIN + LEFT_LINE_BORDER - 4;

  if (c->type == BIG_CURSOR) {
    y = l->dim.y - l->dim.min_y + 4;
    height = 4 * LINE_SPACING - 8;
  } else {
    elem *e = &l->u.music.data[c->chord].u.chord.data[c->note];
    y = l->dim.y - l->dim.min_y + 3 * LINE_SPACING - LINE_SPACING/2 - 2;
    switch (e->type) {
      case NOTE:
        pos = e->u.note.pitch;
        y += -pos * LINE_SPACING/2;
        break;
      case SILENCE:
        pos = e->u.silence.position;
        y += -pos * LINE_SPACING/2;
        break;
    }
    height = 12;

#if 0
    /* if one note of the chord has a diese or bemol or becarre don't draw */
    /* if there is a silence != NOIRE && != CROCHE, don't draw either */
    has_to_draw = 1;
    for (i = 0; has_to_draw &&
                i < CHORD_SIZE((doc *)c->doc, c->line, c->chord); i++) {
      e = &l->u.music.data[c->chord].u.chord.data[i];
      switch (e->type) {
        case NOTE:
          has_to_draw = e->u.note.alteration == NORMAL;
          break;
        case SILENCE:
          has_to_draw = e->u.silence.duration == NOIRE ||
                        e->u.silence.duration == CROCHE;
          break;
      }
    }
#endif

    /* we have to draw a tiny line if the cursor is close enough
     * to the chord
     * 3-l->u.music.data[c->chord].dim.min_x is the distance of
     * the cursor from the chord
     */
    has_to_draw = 3-l->u.music.data[c->chord].dim.min_x <= MINILINE_WIDTH/2;
  }

  gtk_set_color(COLOR_WHITE);
  gtk_wrapper_draw_line(((doc *)c->doc)->drawing_area, x, y, x, y+height,
                        ((doc *)c->doc)->zoom);
  gtk_set_color(COLOR_BLACK);

  /* HAHA! that's the fun part of gcomposer (ok, one of): we redraw the
   * underlying stuff behind the cursor, which can only be lines!!
   */
  if (c->type == BIG_CURSOR) {
    doc *d = c->doc;
    y = l->dim.y - l->dim.min_y;
    gtk_wrapper_draw_line(d->drawing_area, x, y, x+1, y, d->zoom);
    gtk_wrapper_draw_line(d->drawing_area, x, y+LINE_SPACING,
                          x+1, y+LINE_SPACING, d->zoom);
    gtk_wrapper_draw_line(d->drawing_area, x, y+LINE_SPACING*2,
                          x+1, y+LINE_SPACING*2, d->zoom);
    gtk_wrapper_draw_line(d->drawing_area, x, y+LINE_SPACING*3,
                          x+1, y+LINE_SPACING*3, d->zoom);
    gtk_wrapper_draw_line(d->drawing_area, x, y+LINE_SPACING*4,
                          x+1, y+LINE_SPACING*4, d->zoom);
  } else {
    doc *d = c->doc;
    /* odd stuff: center line */
    if (!(pos&1)) {
      if ((pos <= 6 && pos >= -2) || has_to_draw) {
        y = l->dim.y - l->dim.min_y - ((pos-8)/2 + 1) * LINE_SPACING;
        gtk_wrapper_draw_line(d->drawing_area, x, y, x+1, y, d->zoom);
      }
    } else {
      /* up part? */
      if (pos >= 7) {
        elem *last;
        int last_pos;
        if (has_to_draw || pos == 7) {
          y = l->dim.y - l->dim.min_y - ((pos-9)/2 + 1) * LINE_SPACING;
          gtk_wrapper_draw_line(d->drawing_area, x, y, x+1, y, d->zoom);
        }
        /* if we are not the last note, an upper line must be drawn
         * but only if the last note is *really* higher
         */
        last = &l->u.music.data[c->chord].u.chord.data
                   [l->u.music.data[c->chord].u.chord.size-1];
        switch (last->type) {
          case NOTE: last_pos = last->u.note.pitch; break;
          case SILENCE: last_pos = last->u.silence.position; break;
        }
        if (pos < last_pos && has_to_draw)
          gtk_wrapper_draw_line(d->drawing_area, x, y - LINE_SPACING, x+1,
                                y - LINE_SPACING, d->zoom);
      }
      /* down part? */
      else if (pos <= -3) {
        elem *first;
        int first_pos;
        if (has_to_draw || pos == -3) {
          y = l->dim.y - l->dim.min_y - ((pos-9)/2 + 1) * LINE_SPACING;
          gtk_wrapper_draw_line(d->drawing_area, x, y - LINE_SPACING, x+1,
                                y - LINE_SPACING, d->zoom);
        }
        /* if we are not the first note, a lower line must be drawn
         * but only if the first note is *really* lower
         */
        first = &l->u.music.data[c->chord].u.chord.data[0];
        switch (first->type) {
          case NOTE: first_pos = first->u.note.pitch; break;
          case SILENCE: first_pos = first->u.silence.position; break;
        }
        if (pos > first_pos && has_to_draw)
          gtk_wrapper_draw_line(d->drawing_area, x, y, x+1, y, d->zoom);
      }
      /* middle part */
      else {
        /* not up nor high? always draw both lines */
        y = l->dim.y - l->dim.min_y - ((pos-9)/2 + 1) * LINE_SPACING;
        gtk_wrapper_draw_line(d->drawing_area, x, y, x+1, y, d->zoom);
        gtk_wrapper_draw_line(d->drawing_area, x, y - LINE_SPACING, x+1,
                              y - LINE_SPACING, d->zoom);
      }
    }
  }
}

static void clear_cursor_on_text_line(cursor *c)
{
  int x;
  int y;
  int height;
  line *l;

  ensure_line_positioned(c->doc, c->line);

  l = ((doc *)c->doc)->data + c->line;
  if (c->chord == 0)
    x = LEFT_MARGIN + TEXT_LEFT_BORDER - 3;
  else
    x = LEFT_MARGIN + l->dim.width - 3;

#if 0
  y = l->dim.y - l->dim.min_y - 2;
  height = l->dim.max_y - l->dim.min_y + 1 + 4;
#endif
  y = l->dim.y + TEXT_UPPER_MARGIN;
  height = l->dim.height - TEXT_UPPER_MARGIN - TEXT_LOWER_MARGIN;

  gtk_set_color(COLOR_WHITE);
  gtk_wrapper_draw_line(((doc *)c->doc)->drawing_area, x, y, x, y+height,
                        ((doc *)c->doc)->zoom);
  gtk_set_color(COLOR_BLACK);
}

void clear_cursor(cursor *c)
{
  switch (LINE_TYPE((doc *)c->doc, c->line)) {
    case LINE_MUSIC: return clear_cursor_on_music_line(c);
    case LINE_TEXT: return clear_cursor_on_text_line(c);
  }
}

static void draw_cursor_on_music_line(cursor *c)
{
  int x;
  int y;
  int height;
  line *l;

  dprintf("draw_cursor\n");

  if (c->hidden_level) return;
  if (!c->state) return;

  ensure_line_positioned(c->doc, c->line);

  l = ((doc *)c->doc)->data + c->line;
  if (c->chord < l->u.music.size)
    x = LEFT_MARGIN +
        l->u.music.data[c->chord].dim.x - 3;
  else if (l->u.music.size)
    x = LEFT_MARGIN +
        l->u.music.data[l->u.music.size-1].dim.x +
        l->u.music.data[l->u.music.size-1].dim.width - 3;
  else
    x = LEFT_MARGIN + LEFT_LINE_BORDER - 4;

  if (c->type == BIG_CURSOR) {
    y = l->dim.y - l->dim.min_y + 4;
    height = 4 * LINE_SPACING - 8;
  } else {
    elem *e = &l->u.music.data[c->chord].u.chord.data[c->note];
    y = l->dim.y - l->dim.min_y + 3 * LINE_SPACING - LINE_SPACING/2 - 2;
    switch (e->type) {
      case NOTE: y += - e->u.note.pitch * LINE_SPACING/2; break;
      case SILENCE: y += -e->u.silence.position * LINE_SPACING/2; break;
    }
    height = 12;
  }

  gtk_wrapper_draw_line(((doc *)c->doc)->drawing_area, x, y, x, y+height,
                        ((doc *)c->doc)->zoom);
}

static void draw_cursor_on_text_line(cursor *c)
{
  int x;
  int y;
  int height;
  line *l;

  ensure_line_positioned(c->doc, c->line);

  l = ((doc *)c->doc)->data + c->line;
  if (c->chord == 0)
    x = LEFT_MARGIN + TEXT_LEFT_BORDER - 3;
  else
    x = LEFT_MARGIN + l->dim.width - 3;

#if 0
  y = l->dim.y - l->dim.min_y - 2;
  height = l->dim.max_y - l->dim.min_y + 1 + 4;
#endif
  y = l->dim.y + TEXT_UPPER_MARGIN;
  height = l->dim.height - TEXT_UPPER_MARGIN - TEXT_LOWER_MARGIN;

  gtk_wrapper_draw_line(((doc *)c->doc)->drawing_area, x, y, x, y+height,
                        ((doc *)c->doc)->zoom);
}

void draw_cursor(cursor *c)
{
  switch (LINE_TYPE((doc *)c->doc, c->line)) {
    case LINE_MUSIC: return draw_cursor_on_music_line(c);
    case LINE_TEXT: return draw_cursor_on_text_line(c);
  }
}

static void show_cursor_on_music_line(cursor *c)
{
  /* ensure that the cursor is visible (scroll if necessary) */
  int x;
  int y;
  int height;
  line *l;

#if 0
  /* if the cursor is disabled, don't show it */
  if (c->hidden_level) return;
  /* no! if one calls show_cursor, one wants to go to the cursor,
   * we don't decide here
   */
#endif

  /* be sure the cursor's line position is valid and the cursor
   * will be visible visible when we call gtk_wrapper_drawing_area_view
   */
  ensure_line_positioned(c->doc, c->line);

#if 0
  /* allow shrink */
  update_scrollbars(c->doc, 1);
#endif

  l = ((doc *)c->doc)->data + c->line;
  x = LEFT_MARGIN-4;
  x += LEFT_LINE_BORDER;
  if (c->chord < l->u.music.size) {
    x += l->u.music.data[c->chord].dim.x;
  } else {
    if (l->u.music.size)
      x += l->u.music.data[l->u.music.size-1].dim.x +
           l->u.music.data[l->u.music.size-1].dim.width;
  }

  dprintf("showcursor line %d y of line %d invalid %d\n",
          c->line, l->dim.y, l->next_invalid);
  y = l->dim.y /* UPPER_MARGIN + c->line * LINE_HEIGHT */;

  height = l->dim.height;

  y -= 20;
  height += 40;

  dprintf("show cursor %d %d (line h=%d)\n", x-20, y, l->dim.height);
  gtk_wrapper_drawing_area_view(((doc *)c->doc)->drawing_area,
                                x-20, y, 80, height,
                                ((doc *)c->doc)->zoom);

  /* the scrollbar may have to be changed */
  update_horizontal_scrollbar((doc *)c->doc);
  /* in case the scrollbar changed the cursor may not be visible,
   * so redo it (yeah, really ugly)
   */
  gtk_wrapper_drawing_area_view(((doc *)c->doc)->drawing_area,
                                x-20, y, 80, height,
                                ((doc *)c->doc)->zoom);
}

static void show_cursor_on_text_line(cursor *c)
{
  int x;
  int y;
  int height;
  line *l;

  ensure_line_positioned(c->doc, c->line);

  l = ((doc *)c->doc)->data + c->line;
  if (c->chord == 0)
    x = LEFT_MARGIN + TEXT_LEFT_BORDER - 3;
  else
    x = LEFT_MARGIN + l->dim.width - 3;

#if 0
  y = l->dim.y - l->dim.min_y - 2;
  height = l->dim.max_y - l->dim.min_y + 1 + 4;
#endif
  y = l->dim.y + TEXT_UPPER_MARGIN;
  height = l->dim.height - TEXT_UPPER_MARGIN - TEXT_LOWER_MARGIN;

  y -= 20;
  height += 40;

  gtk_wrapper_drawing_area_view(((doc *)c->doc)->drawing_area,
                                x-20, y, 80, height,
                                ((doc *)c->doc)->zoom);

  /* the scrollbar may have to be changed */
  update_horizontal_scrollbar((doc *)c->doc);
  /* in case the scrollbar changed the cursor may not be visible,
   * so redo it (yeah, really ugly)
   */
  gtk_wrapper_drawing_area_view(((doc *)c->doc)->drawing_area,
                                x-20, y, 80, height,
                                ((doc *)c->doc)->zoom);
}

void show_cursor(cursor *c)
{
  dprintf("show_cursor\n");
  switch (LINE_TYPE((doc *)c->doc, c->line)) {
    case LINE_MUSIC: return show_cursor_on_music_line(c);
    case LINE_TEXT: return show_cursor_on_text_line(c);
  }
}

void show_chord(doc *d, int line_number, int chord_number)
{
  /* ensure that the chord is visible (scroll if necessary) */
  int x;
  int y;
  int height;
  int width;
  line *l;
  chord *c;

  ensure_line_positioned(d, line_number);

  l = d->data + line_number;
  x = LEFT_MARGIN-4;
  x += LEFT_LINE_BORDER;
  if (chord_number < l->u.music.size) {
    x += l->u.music.data[chord_number].dim.x;
  } else {
    if (l->u.music.size)
      x += l->u.music.data[l->u.music.size-1].dim.x +
           l->u.music.data[l->u.music.size-1].dim.width;
  }

  y = l->dim.y; /* UPPER_MARGIN + line_number * LINE_HEIGHT */;

  height = l->dim.height;

  if (chord_number < l->u.music.size) {
    c = &l->u.music.data[chord_number];
    width = c->dim.max_x - c->dim.min_x + 1;
  } else
    width = 0;

  y -= 10;
  height += 20;

  x -= 10;
  width += 20;
  if (width < 60) width = 60;

  gtk_wrapper_drawing_area_view(d->drawing_area,
                                x, y, width, height,
                                d->zoom);

  /* the scrollbar may have to be changed */
  update_horizontal_scrollbar(d);
  /* in case the scrollbar changed the cursor may not be visible,
   * so redo it (yeah, really ugly)
   */
  gtk_wrapper_drawing_area_view(d->drawing_area,
                                x, y, width, height,
                                d->zoom);
}
