/* gcomposer: a little tool to write music.
 * http://sed.free.fr/gcomposer
 * Released in the public domain (freedom has no license).
 */

#include "file.h"

#include <stdio.h>
#include <errno.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <stdlib.h>

#include "error.h"
#include "dim.h"

void remove_space(FILE *f)
{
  while (1) {
    int c = getc(f);
    if (c == EOF) return;
    if (!isspace(c)) {
      ungetc(c, f);
      return;
    }
  }
}

int load(doc *d, char *file)
{
  FILE *f = fopen(file, "r");
  doc ret;
  int major_version;
  int minor_version;
  int chord_viewed;
  int fret;
  int one_music_line_loaded = 0;

  if (!f) {
    set_error(FILE_ERROR, FILE_OPEN_ERROR, file, errno);
    return 0;
  }

  memset(&ret, 0, sizeof(doc));

  /* first stuff of the file: [gcomposer MAJOR_VERSION.MIN_VERSION] */
  remove_space(f);
  if (fscanf(f, "[") == EOF) goto read_error;
  remove_space(f);
  if (fscanf(f, "gcomposer") == EOF) goto read_error;
  remove_space(f);
  fret = fscanf(f, "%d.%d", &major_version, &minor_version);
  if (fret == EOF) goto read_error;
  if (fret != 2) goto load_error;
  remove_space(f);
  if (fscanf(f, "]") == EOF) goto read_error;
  if (major_version > 3 || minor_version != 0) goto load_error;

  /* now the stuff */
  chord_viewed = 0;
  while (1) {
    char command[128];
    int i;
    int c;

    remove_space(f);

    i = 0;
    /* get the command */
    do {
      c = getc(f);
      if (c != EOF && !isspace(c) && c!='(') { command[i] = c; i++; }
    } while (c != EOF && !isspace(c) && c != '(' && i != 128);
    /* we go out of the loop if no more item to parse */
    if (i == 0) break;
    if (i == 128) goto load_error;
    command[i] = 0;
    if (c == '(') ungetc(c, f);

    remove_space(f);

    if (!strcmp(command, "tempo")) {
      if (chord_viewed) goto load_error;
      fret = fscanf(f, "%d", &ret.tempo);
      if (fret == EOF) goto read_error;
      if (fret != 1) goto load_error;
      if (ret.tempo < 0 || ret.tempo > 1000) goto load_error;
    } else if (!strcmp(command, "textline")) {
      int len;
      int pos;
      char *t;
      if (chord_viewed) goto load_error;
      fret = fscanf(f, "(%d)", &len);
      if (fret == EOF) goto read_error;
      if (fret != 1) goto load_error;
      if (!ret.size) {
        if (!doc_insert_line(&ret, LINE_TEXT, -1)) goto doc_creation_error;
      } else {
        if (!doc_insert_line(&ret, LINE_TEXT, ret.size-1))
          goto doc_creation_error;
      }
      LINE_TEXT_DATA(&ret, ret.size-1) = t = malloc(len+1);
      if (!t) {
        set_error(OOM);
        goto doc_creation_error;
      }
      for (pos = 0; pos < len; pos++) {
        int c = getc(f);
        if (c == EOF) goto load_error; /* should it be read_error? */
        /* no newline allowed in text */
        if (c == '\n') goto load_error;
        t[pos] = c;
      }
      t[pos] = 0;
    } else if (!strcmp(command, "line")) {
      if (chord_viewed) goto load_error;
      one_music_line_loaded = 1;
      if (!ret.size) {
        if (!doc_insert_line(&ret, LINE_MUSIC, -1)) goto doc_creation_error;
      } else {
        if (!doc_insert_line(&ret, LINE_MUSIC, ret.size-1))
          goto doc_creation_error;
      }
    } else if (!strcmp(command, "chord")) {
      if (chord_viewed) goto load_error;
      chord_viewed=1;
    } else if (!strcmp(command, "note")) {
      int c;
      int pitch;
      int alteration;
      int duration;
      int nb_dots;
      int head_delta = 0;
      int alteration_delta = 0;
      int dots_delta = 0;
      int stem_down = 0;
      if (chord_viewed) {
        if (!ret.size || ret.data[ret.size-1].type != LINE_MUSIC)
          goto load_error;
        if (!doc_insert_chord(&ret, ret.size-1, -1, NOTES_CHORD))
          goto doc_creation_error;
      }
      chord_viewed = 0;
      remove_space(f);
      if ((c=getc(f)) == '(') {
        do {
          remove_space(f);
          c = getc(f);
          /* this code accepts more than one 'n' and more than one 'a'
           * (and 'd')
           * the last one is taken into account, the others are forgotten
           * we could be more anal, but hey, come on...
           */
          if (c == 'n') {
            remove_space(f);
            c = getc(f);
            if (c != ':') goto load_error;
            fret = fscanf(f, "%d", &head_delta);
            if (fret == EOF) goto read_error;
            if (fret != 1) goto load_error;
            if (head_delta < MIN_NOTE_HEAD_DELTA)
              head_delta = MIN_NOTE_HEAD_DELTA;
            if (head_delta > MAX_NOTE_HEAD_DELTA)
              head_delta = MAX_NOTE_HEAD_DELTA;
          } else if (c == 'a') {
            remove_space(f);
            c = getc(f);
            if (c != ':') goto load_error;
            fret = fscanf(f, "%d", &alteration_delta);
            if (fret == EOF) goto read_error;
            if (fret != 1) goto load_error;
            if (alteration_delta < MIN_ALTERATION_DELTA)
              alteration_delta = MIN_ALTERATION_DELTA;
            if (alteration_delta > MAX_ALTERATION_DELTA)
              alteration_delta = MAX_ALTERATION_DELTA;
          } else if (c == 'd') {
            remove_space(f);
            c = getc(f);
            if (c != ':') goto load_error;
            fret = fscanf(f, "%d", &dots_delta);
            if (fret == EOF) goto read_error;
            if (fret != 1) goto load_error;
            if (dots_delta < MIN_DOTS_DELTA)
              dots_delta = MIN_DOTS_DELTA;
            if (dots_delta > MAX_DOTS_DELTA)
              dots_delta = MAX_DOTS_DELTA;
          } else if (c == 's') {
            stem_down = 1;
          } else goto load_error;
          remove_space(f);
          c = getc(f);
        } while (c == ',');
        if (c != ')') goto load_error;
      } else /* c != '(' */
        ungetc(c, f);
      fret = fscanf(f, "%d %d %d %d", &pitch, &alteration,
                    &duration, &nb_dots);
      if (fret == EOF) goto read_error;
      if (fret != 4) goto load_error;
      if (pitch < STAFF_NOTE_MIN || pitch > STAFF_NOTE_MAX) goto load_error;
      /* alteration is note_alteration type */
      if (alteration < 0 || alteration > 3) goto load_error;
      /* duration is note_duration type */
      if (duration < 0 || duration > 7) goto load_error;
      if (nb_dots < 0 || nb_dots > 3) goto load_error;
      /* current stuff must be a chord, it must be of good type */
      if (!ret.size || ret.data[ret.size-1].type != LINE_MUSIC ||
          !ret.data[ret.size-1].u.music.size ||
          ret.data[ret.size-1].u.music.data[
              ret.data[ret.size-1].u.music.size-1].type != NOTES_CHORD)
        goto load_error;
      if (!doc_insert_note(&ret, ret.size-1,
                           ret.data[ret.size-1].u.music.size-1,
                           NOTE, pitch, alteration, duration, nb_dots,
                           head_delta, alteration_delta, dots_delta,
                           stem_down))
        goto doc_creation_error;
    } else if (!strcmp(command, "silence")) {
      int position;
      int duration;
      int nb_dots;
      if (chord_viewed) {
        if (!ret.size || ret.data[ret.size-1].type != LINE_MUSIC)
          goto load_error;
        if (!doc_insert_chord(&ret, ret.size-1, -1, NOTES_CHORD))
          goto doc_creation_error;
      }
      chord_viewed = 0;
      fret = fscanf(f, "%d %d %d", &position, &duration, &nb_dots);
      if (fret == EOF) goto read_error;
      if (fret != 3) goto load_error;
      if (position < STAFF_NOTE_MIN || position > STAFF_NOTE_MAX)
        goto load_error;
      /* duration is note_duration type */
      if (duration < 0 || duration > 7) goto load_error;
      if (nb_dots < 0 || nb_dots > 3) goto load_error;
      /* current stuff must be a chord, it must be of good type */
      if (!ret.size || ret.data[ret.size-1].type != LINE_MUSIC ||
          !ret.data[ret.size-1].u.music.size ||
          ret.data[ret.size-1].u.music.data[
              ret.data[ret.size-1].u.music.size-1].type != NOTES_CHORD)
        goto load_error;
      if (!doc_insert_note(&ret, ret.size-1,
                           ret.data[ret.size-1].u.music.size-1,
                           SILENCE, position, duration, nb_dots))
        goto doc_creation_error;
    } else if (!strcmp(command, "mesure_break")) {
      if (chord_viewed) goto load_error;
      if (!ret.size || ret.data[ret.size-1].type != LINE_MUSIC)
        goto load_error;
      if (!doc_insert_chord(&ret, ret.size-1, -1, MESURE_BREAK))
        goto doc_creation_error;
    } else if (!strcmp(command, "tonalite")) {
      int position;
      int alteration;
      if (chord_viewed) goto load_error;
      fret = fscanf(f, "%d %d", &position, &alteration);
      if (fret == EOF) goto read_error;
      if (fret != 2) goto load_error;
      if (position < STAFF_NOTE_MIN || position > STAFF_NOTE_MAX)
        goto load_error;
      /* alteration is note_alteration type */
      if (alteration < 0 || alteration > 3) goto load_error;
      if (!ret.size || ret.data[ret.size-1].type != LINE_MUSIC)
        goto load_error;
      if (!doc_insert_chord(&ret, ret.size-1, -1, TONALITE))
        goto doc_creation_error;
      TONALITE_POSITION(&ret, ret.size-1,
                        ret.data[ret.size-1].u.music.size-1) = position;
      TONALITE_ALTERATION(&ret, ret.size-1,
                          ret.data[ret.size-1].u.music.size-1) = alteration;
    } else if (!strcmp(command, "clef")) {
      char type;
      int position;
      if (chord_viewed) goto load_error;
      fret = fscanf(f, "%c %d", &type, &position);
      if (fret == EOF) goto read_error;
      if (fret != 2) goto load_error;
      if (type != 'G' && type != 'F' && type != 'C' &&
          type != 'a' && type != 'b') goto load_error;
      if (position < CLEF_POS_MIN || position > CLEF_POS_MAX) goto load_error;
      if (!ret.size || ret.data[ret.size-1].type != LINE_MUSIC)
        goto load_error;
      if (!doc_insert_chord(&ret, ret.size-1, -1, CLEF))
        goto doc_creation_error;
      switch (type) {
        case 'G':
          CLEF_TYPE(&ret, ret.size-1,
                    ret.data[ret.size-1].u.music.size-1) = CLEF_G;
          break;
        case 'F':
          CLEF_TYPE(&ret, ret.size-1,
                    ret.data[ret.size-1].u.music.size-1) = CLEF_F;
          break;
        case 'C':
          CLEF_TYPE(&ret, ret.size-1,
                    ret.data[ret.size-1].u.music.size-1) = CLEF_C;
          break;
        case 'a':
          CLEF_TYPE(&ret, ret.size-1,
                    ret.data[ret.size-1].u.music.size-1) = CLEF_G_8va;
          break;
        case 'b':
          CLEF_TYPE(&ret, ret.size-1,
                    ret.data[ret.size-1].u.music.size-1) = CLEF_G_8vb;
          break;
      }
      CLEF_POSITION(&ret, ret.size-1,
                    ret.data[ret.size-1].u.music.size-1) = position;
    } else goto load_error;
  }

  if (fclose(f) < 0) {
    set_error(FILE_ERROR, FILE_CLOSE_ERROR, file, errno);
    return 0;
  }
  f = NULL;

  /* sanity checks */
  /* a music line must be here */
  if (!one_music_line_loaded) goto load_error;
  /* we don't allow empty chord, even at the end of file */
  if (chord_viewed) goto load_error;

  destroy_doc(d);
  d->data = ret.data;
  d->size = ret.size;
  d->maxsize = ret.maxsize;
  d->cursor.type = BIG_CURSOR;
  d->cursor.line = 0;
  d->cursor.chord = 0;
  d->tempo = ret.tempo;
  d->sel.start_line = -1;

  return 1;

read_error:
  set_error(FILE_ERROR, FILE_READ_ERROR, file, errno);
  destroy_doc(&ret);
  fclose(f);
  return 0;

load_error:
  if (f) fclose(f);
  destroy_doc(&ret);
  set_error(FILE_ERROR, FILE_LOAD_ERROR, file, 0); /* errno is not used */
  return 0;

doc_creation_error:
  if (f) fclose(f);
  destroy_doc(&ret);
  return 0;
}

int save(doc *d, char *file)
{
  int i;
  FILE *f = fopen(file, "w");
  if (!f) {
    set_error(FILE_ERROR, FILE_OPEN_ERROR, file, errno);
    return 0;
  }

  if (fprintf(f, "[gcomposer 3.0]\n") < 0) goto write_error;

  if (fprintf(f, "tempo %d\n", d->tempo) < 0) goto write_error;

  for (i = 0; i < d->size; i++) {
    int j;
    line *l = d->data+i;

    /* text line */
    if (l->type == LINE_TEXT) {
      if (fprintf(f, "textline(%d)%s\n",
                  strlen(LINE_TEXT_DATA(d, i)),
                  LINE_TEXT_DATA(d, i)) < 0)
        goto write_error;
      continue;
    }

    /* music line */
    if (fprintf(f, "line\n") < 0) goto write_error;

    for (j = 0; j < l->u.music.size; j++) {
      chord *c = l->u.music.data+j;

      if (fprintf(f, "  ") < 0) goto write_error;

      switch (c->type) {
        case NOTES_CHORD: {
          int k;

          if (fprintf(f, "chord\n") < 0) goto write_error;

          for (k = 0; k < c->u.chord.size; k++) {
            elem *e = c->u.chord.data+k;

            if (fprintf(f, "    ") < 0) goto write_error;

            switch (e->type) {
              case NOTE:
                if (fprintf(f, "note") < 0) goto write_error;
                if (e->u.note.head_delta || e->u.note.alteration_delta ||
                    e->u.note.dots_delta) {
                  if (fprintf(f, "(") < 0) goto write_error;
                  if (e->u.note.head_delta) {
                    if (fprintf(f, "n:%d", e->u.note.head_delta) < 0)
                      goto write_error;
                    if (e->u.note.alteration_delta || e->u.note.dots_delta ||
                        e->u.note.stem_down)
                      if (fprintf(f, ", ") < 0) goto write_error;
                  }
                  if (e->u.note.alteration_delta) {
                    if (fprintf(f, "a:%d", e->u.note.alteration_delta) < 0)
                      goto write_error;
                    if (e->u.note.dots_delta || e->u.note.stem_down)
                      if (fprintf(f, ", ") < 0) goto write_error;
                  }
                  if (e->u.note.dots_delta) {
                    if (fprintf(f, "d:%d", e->u.note.dots_delta) < 0)
                      goto write_error;
                    if (e->u.note.stem_down)
                      if (fprintf(f, ", ") < 0) goto write_error;
                  }
                  if (e->u.note.stem_down)
                    if (fprintf(f, "s") < 0) goto write_error;
                  if (fprintf(f, ")") < 0) goto write_error;
                }
                if (fprintf(f, " %d %d %d %d",
                            e->u.note.pitch,
                            e->u.note.alteration,
                            e->u.note.duration,
                            e->u.note.nb_dots) < 0) goto write_error;
                break;
              case SILENCE:
                if (fprintf(f, "silence %d %d %d",
                            e->u.silence.position,
                            e->u.silence.duration,
                            e->u.silence.nb_dots) < 0) goto write_error;
                break;
            }

            if (fprintf(f, "\n") < 0) goto write_error;
          }
          break;
        }
        case MESURE_BREAK:
          if (fprintf(f, "mesure_break\n") < 0) goto write_error;
          break;
        case TONALITE:
          if (fprintf(f, "tonalite %d %d\n",
                      c->u.tonalite.position,
                      c->u.tonalite.alteration) < 0) goto write_error;
          break;
        case CLEF:
          if (fprintf(f, "clef %c %d\n",
                      c->u.clef.type == CLEF_G ? 'G' :
                        c->u.clef.type == CLEF_F ? 'F' :
                        c->u.clef.type == CLEF_C ? 'C' :
                        c->u.clef.type == CLEF_G_8va ? 'a' : 'b',
                      c->u.clef.position) < 0) goto write_error;
          break;
      }
    }
  }

  if (fclose(f) < 0) {
    set_error(FILE_ERROR, FILE_CLOSE_ERROR, file, errno);
    return 0;
  }

  return 1;

write_error:
  set_error(FILE_ERROR, FILE_WRITE_ERROR, file, errno);
  fclose(f);
  return 0;
}

int file_exist(char *filename)
{
  struct stat buf;
  return stat(filename, &buf) == 0;
}
