/*
 * l: a language for cgscore to describe actions
 * public domain, like cgscore
 */

#include <stdio.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/types.h>
#include <dirent.h>
#include <string.h>
#include <errno.h>

#include "files.h"
#include "parser.h"
#include "list.h"
#include "string.h"
#include "actions.h"

/* barbarian globals */
list include;
list cinclude;
list hinclude;

/* structure:
 * <action_name> <action_argument (-1 if none)>
 * <data> <do> <undo> <redo> <save> <load> <erase>
 * repeated for each action
 */
list actions;

/* structure:
 * <action_name> <position in list 'actions' above>
 */
list registered_actions;

/* -1 error, 0 ok */
int run_file(char *dir, char *file)
{
  char name[FILENAME_MAX];

  if (strlen(dir) + strlen(file) + 2 >= FILENAME_MAX) {
    fprintf(stderr, "%s/%s: filename too long\n", dir, file);
    return -1;
  }

  sprintf(name, "%s/%s", dir, file);

  if (init_file(name) == -1) {
    fprintf(stderr, "%s: error\n", name);
    return -1;
  }

  reset_parse();

  if (yyparse() == 1) {
    fprintf(stderr, "%s: parse error\n", name);
    return -1;
  }

  return 0;
}

/* -1 error, 0 ok */
int run_dir(char *dir)
{
  DIR *d = opendir(dir);
  struct dirent *f;

  if (!d) {
    perror(dir);
    return -1;
  }

  while ((f = readdir(d))) {
    int l = strlen(f->d_name);
    if (l < 3) continue;
    /* we do files ending with '.l' */
    if (f->d_name[l-2] != '.' || f->d_name[l-1] != 'l') continue;
    if (run_file(dir, f->d_name) == -1) return -1;
  }

  closedir(d);

  if (errno == EBADF)
    return -1;

  return 0;
}

void dump(void)
{
  int i;
  fprintf(stderr, "include:\n");
  for (i=0; i<list_size(include); i++)
    fprintf(stderr, "  %s\n", get_string(list_get_int(include, i)));
  fprintf(stderr, "hinclude:\n");
  for (i=0; i<list_size(hinclude); i++)
    fprintf(stderr, "  %s\n", get_string(list_get_int(hinclude, i)));
  fprintf(stderr, "cinclude:\n");
  for (i=0; i<list_size(cinclude); i++)
    fprintf(stderr, "  %s\n", get_string(list_get_int(cinclude, i)));
  fprintf(stderr, "actions:\n");
  for (i=0; i<list_size(actions); i+=8) {
    fprintf(stderr, "  %s\n", get_string(list_get_int(actions, i)));
    fprintf(stderr, "  args:\n");
    fprintf(stderr, "    %s\n", get_string(list_get_int(actions, i+1)));
    fprintf(stderr, "  data:\n");
    fprintf(stderr, "    %s\n", get_string(list_get_int(actions, i+2)));
    fprintf(stderr, "  do:\n");
    fprintf(stderr, "    %s\n", get_string(list_get_int(actions, i+3)));
    fprintf(stderr, "  undo:\n");
    fprintf(stderr, "    %s\n", get_string(list_get_int(actions, i+4)));
    fprintf(stderr, "  redo:\n");
    fprintf(stderr, "    %s\n", get_string(list_get_int(actions, i+5)));
    fprintf(stderr, "  save:\n");
    fprintf(stderr, "    %s\n", get_string(list_get_int(actions, i+6)));
    fprintf(stderr, "  load:\n");
    fprintf(stderr, "    %s\n", get_string(list_get_int(actions, i+7)));
  }
}

int main(int n, char **v)
{
  int i;
  char *hout = 0;
  char *henum = 0;

  if (n < 2) {
arg_err:
    fprintf(stderr, "usage: %s [options] <dir 1> .. <dir n>\n", v[0]);
    fprintf(stderr, "options:\n");
    fprintf(stderr, "    -hout <file>\n");
    fprintf(stderr, "        dump the .h stuff in <file>\n");
    fprintf(stderr, "    -henum <file>\n");
    fprintf(stderr, "        dump the enum .h stuff in <file>\n");
    return 1;
  }

  /* run through args */
  for (i = 1; i < n; i++) {
    struct stat buf;

    if (!strcmp(v[i], "-hout")) {
      if (hout || i == n - 1) goto arg_err;
      v[i] = 0;
      i++;
      hout = v[i];
      v[i] = 0;
      continue;
    }

    if (!strcmp(v[i], "-henum")) {
      if (henum || i == n - 1) goto arg_err;
      v[i] = 0;
      i++;
      henum = v[i];
      v[i] = 0;
      continue;
    }

    /* not an arg, so must be a dir */
    if (stat(v[i], &buf) == -1) {
      perror(v[i]);
      return 1;
    }
    if (!(buf.st_mode & S_IFDIR)) {
      fprintf(stderr, "%s is not a directory\n", v[i]);
      return 1;
    }
  }

  if (hout) { extern char *H; H = hout; }
  if (henum) { extern char *N; N = henum; }

  /* create structures */
  include = new_list();
  cinclude = new_list();
  hinclude = new_list();
  actions = new_list();
  registered_actions = new_list();
  if (include == -1 || cinclude == -1 || hinclude == -1 ||
      actions == -1 || registered_actions == -1) {
    fprintf(stderr, "oom?\n");
    return 1;
  }

  /* get registered actions */
  if (get_registered_actions() == -1)
    return 1;

  /* run into directories */
  for (i = 1; i < n; i++) {
    if (v[i] == 0) continue;

    if (run_dir(v[i]) == -1) {
      fprintf(stderr, "error in directory %s\n", v[i]);
      return 1;
    }
  }

  /* save actions */
  if (store_actions() == -1) {
    fprintf(stderr, "error storing actions\n");
    return 1;
  }

  /* create actions.h and actions.c */
  if (create_actions() == -1) {
    fprintf(stderr, "error creating actions.h and/or actions.c\n");
    return 1;
  }

  //dump();

  return 0;
}
