/*
 * m: a language for cgscore to describe states
 * public domain, like cgscore
 */

#include <string.h>
#include <stdio.h>
#include <sys/types.h>
#include <dirent.h>
#include <errno.h>
#include <sys/stat.h>
#include <unistd.h>

#include "list.h"
#include "files.h"
#include "parser.h"
#include "states.h"

/* barbarian globals */
list include;
list cinclude;
list hinclude;

/* structure:
 *   <state 1> ... <state n>
 * each state is a list containing:
 *   <identifier> <filename of state> <arguments (-1 if no arguments)>
 *   <data> <enter> <leave> <actions>
 * actions is a flat list:
 *   <identifier 1> <arguments 1 (-1 if no arguments)> <code 1> ...
 *     <identifier m> <arguments m (-1 if no arguments)> <code m>
 *   arguments X is a list:
 *     <arguments string> <argument 1 start offset> <argument 1 end offset>
 *       <argument 1 name start offset> <argument 1 name end offset>
 *       <argument m start offset> <argument m end offset>
 *       <argument m name start offset> <argument m name end offset>
 *   code X is a list:
 *     <code string> <call 1 start offset> <call 1 end offset>
 *       <call m start offset> <call m end offset>
 */
list states;

/* other ugly globals */
int only_c;
int verbose;
int split;

/* -1 error, 0 ok */
int run_file(char *dir, char *file)
{
  char name[FILENAME_MAX];

  if (verbose) fprintf(stderr, ".. processing %s/%s...\n", dir, file);

  if (strlen(dir) + strlen(file) + 2 >= FILENAME_MAX) {
    fprintf(stderr, "%s/%s: filename too long\n", dir, file);
    return -1;
  }

  sprintf(name, "%s/%s", dir, file);

  if (init_file(name) == -1) {
    fprintf(stderr, "%s: error\n", name);
    return -1;
  }

  reset_parse();

  if (yyparse() == 1) {
    fprintf(stderr, "%s: parse error\n", name);
    return -1;
  }

  return 0;
}

/* -1 error, 0 ok */
int run_dir(char *dir)
{
  DIR *d = opendir(dir);
  struct dirent *f;

  if (!d) {
    perror(dir);
    return -1;
  }

  while ((f = readdir(d))) {
    int l = strlen(f->d_name);
    if (l < 3) continue;
    /* we do files ending with '.m' */
    if (f->d_name[l-2] != '.' || f->d_name[l-1] != 'm') continue;
    if (run_file(dir, f->d_name) == -1) return -1;
  }

  closedir(d);

  if (errno == EBADF)
    return -1;

  return 0;
}

int main(int n, char **v)
{
  int i;

  if (n < 2) {
    fprintf(stderr, "usage: %s [options] <dir 1> .. <dir n>\n", v[0]);
    fprintf(stderr, "options:\n");
    fprintf(stderr, "    -only-c\n");
    fprintf(stderr, "        don't generate the .h\n");
    fprintf(stderr, "    -v\n");
    fprintf(stderr, "        set verbose mode\n");
    fprintf(stderr, "    -h | --help\n");
    fprintf(stderr, "        show this help message\n");
    fprintf(stderr, "    -hout <file>\n");
    fprintf(stderr, "        dump the .h stuff in <file>\n");
    fprintf(stderr, "    -split\n");
    fprintf(stderr,
      "        dump functions of a state in <file>.c instead of states.c\n");
    return 1;
  }

  /* run through args */
  for (i = 1; i < n; i++) {
    struct stat buf;

    /* check args */
    if (!strcmp(v[i], "-only-c")) {
      only_c = 1;
      v[i] = 0;
      continue;
    }

    if (!strcmp(v[i], "-hout")) {
      if (i == n-1 || hout) {
        /* barbarian hack - the first time I do it! very nice feeling */
        main(1, v);
        return 1;
      }
      v[i] = 0;
      i++;
      hout = v[i];
      v[i] = 0;
      continue;
    }

    if (!strcmp(v[i], "-split")) {
      split = 1;
      v[i] = 0;
      continue;
    }

    if (!strcmp(v[i], "-v")) {
      verbose = 1;
      v[i] = 0;
      continue;
    }

    if (!strcmp(v[i], "-h") || !strcmp(v[i], "--help")) {
      /* barbarian hack - the first time I do it! very nice feeling */
      main(1, v);
      return 0;
    }

    /* not an arg, so it MUST be a directory */
    if (stat(v[i], &buf) == -1) {
      perror(v[i]);
      return 1;
    }
    if (!(buf.st_mode & S_IFDIR)) {
      fprintf(stderr, "%s is not a directory\n", v[i]);
      return 1;
    }
  }

  /* create structures */
  include = new_list();
  cinclude = new_list();
  hinclude = new_list();
  states = new_list();
  if (include == -1 || cinclude == -1 || hinclude == -1 || states == -1) {
    fprintf(stderr, "oom?\n");
    return 1;
  }

  /* run into directories */
  for (i = 1; i < n; i++) {
    if (v[i] == 0) continue;

    if (run_dir(v[i]) == -1) {
      fprintf(stderr, "error in directory %s\n", v[i]);
      return 1;
    }
  }

  if (check_states() == -1) return 1;

  if (create_states() == -1) return 1;

  return 0;
}
