/* gcomposer: a little tool to write music.
 * http://sed.free.fr/gcomposer
 * Released in the public domain (freedom has no license).
 */

/* All over the code I suppose that writing N bytes to midi device will
 *   1 - succeed and write N bytes or,
 *   2 - fail and return -1
 * we might get a:
 *   3 - write less than N bytes
 * this happens with sockets for example, what about midi
 * devices? hopefully it does not happen
 * (anyway it's easy to handle this case, maybe I'll do it before
 * releasing)
 */

#include "midi.h"

#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>

#ifndef WINDOZ

#ifdef ALSA
#include <asoundlib.h>
#endif

#include <linux/soundcard.h>
#include <sys/ioctl.h>

#else /* WINDOZ */

/* Documentation to do MIDI under windoz was taken from:
 * http://www.borg.com/~jglatt/tech/lowmidi.htm
 */
#include <windef.h>
#include <mmreg.h>

/* ugly global, but windoz, so no pb */
static HMIDIOUT midi_windows;
static unsigned long midi_windows_device;

#endif /* WINDOZ */

#include "dprintf.h"

/* needed for the timer, may be removed if another timer mechanism is used */
#include "gtk_wrapper.h"

#ifdef ALSA
static void alsa_play(snd_seq_t *h, int port, int a, int b, int c)
{
  snd_seq_event_t ev;

  snd_seq_ev_clear(&ev);
  snd_seq_ev_set_source(&ev, port);
  snd_seq_ev_set_subs(&ev);
  snd_seq_ev_set_direct(&ev);

  snd_seq_ev_set_noteon(&ev, a & 0x0f, b, c);

  snd_seq_event_output_direct(h, &ev);
}

static void alsa_stop(snd_seq_t *h, int port, int a, int b, int c)
{
  snd_seq_event_t ev;

  snd_seq_ev_clear(&ev);
  snd_seq_ev_set_source(&ev, port);
  snd_seq_ev_set_subs(&ev);
  snd_seq_ev_set_direct(&ev);

  snd_seq_ev_set_noteoff(&ev, a & 0x0f, b, c);

  snd_seq_event_output_direct(h, &ev);
}

static void alsa_inst(snd_seq_t *h, int port, int a, int b)
{
  snd_seq_event_t ev;

  snd_seq_ev_clear(&ev);
  snd_seq_ev_set_source(&ev, port);
  snd_seq_ev_set_subs(&ev);
  snd_seq_ev_set_direct(&ev);

  snd_seq_ev_set_pgmchange(&ev, a & 0x0f, b);

  snd_seq_event_output_direct(h, &ev);
}
#endif /* ALSA */

/* return 0 if ok, -1 if error */
int get_midi_devices(midi *m)
{
#ifdef WINDOZ
  unsigned long nb_devices;
  unsigned long i;
  MIDIOUTCAPS caps;

  m->names = 0;
  m->nb_devices = 0;

  nb_devices = midiOutGetNumDevs();
  if (nb_devices == 0) return -1;

  m->names = malloc(sizeof(char *) * nb_devices);
  if (!m->names) goto err;
  memset(m->names, 0, sizeof(char *) * nb_devices);

  for (i=0; i<nb_devices; i++) {
    if (midiOutGetDevCaps(i, &caps, sizeof(MIDIOUTCAPS))) goto err;
    m->names[i] = strdup(caps.szPname);
    if (!m->names[i]) goto err;
  }

  m->nb_devices = nb_devices;
  return 0;

err:
  if (m->names)
    for (i=0; i<nb_devices; i++)
      if (m->names[i]) free(m->names[i]);
  if (m->names) free(m->names);
  m->names = 0;
  m->nb_devices = 0;
  return -1;
  
#else /* WINDOZ */
  int fd;
  int midi_count;
  int i;
  struct midi_info info;
  char name[12];

  /* try to get stuff by using /dev/sequencer */
  fd = open("/dev/sequencer", O_WRONLY);
  if (fd == -1) goto no_sequencer;

  if (ioctl(fd, SNDCTL_SEQ_NRMIDIS, &midi_count) < 0) {
    close(fd);
    goto no_sequencer;
  }

  m->names = malloc(sizeof(char *) * midi_count);
  if (!m->names) {
    close(fd);
    return -1;
  }
  memset(m->names, 0, sizeof(char *) * midi_count);
  m->nb_devices = midi_count;

  for (i = 0; i < midi_count; i++) {
    info.device = i;
    if (ioctl(fd, SNDCTL_MIDI_INFO, &info) < 0) {
      close(fd);
      goto free_names;
    }
    m->names[i] = strdup(info.name);
    if (!m->names[i]) {
      close(fd);
      goto free_names;
    }
  }

  /* we got it! */
  close(fd);
  return 0;

free_names:
  for (i = 0; i < m->nb_devices; i++) if (m->names[i]) free (m->names[i]);
  free(m->names);
  m->names = NULL;
  m->nb_devices = 0;
  /* fall through, let's try with no sequencer just in case */

no_sequencer:
  /* there is a problem with /dev/sequencer, directly use /dev/midiXX */
  /* we suppose that /dev/midiA exists if and only if /dev/midiB exists
   * for all B < A (so we can stop scanning as soon as open fails)
   */
  for (i = 0; i < 100; i++) {
    sprintf(name, "/dev/midi%2.2d", i);
    fd = open(name, O_WRONLY);
    if (fd == -1) break;
    close(fd);
  }

  /* no device found? */
  if (i == 0) return -1;

  m->names = malloc(sizeof(char *) * i);
  if (!m->names) return -1;
  memset(m->names, 0, sizeof(char *) * i);
  m->nb_devices = i;

  for (i = 0; i < m->nb_devices; i++) {
    char t[15];
    sprintf(t, "Midi device %d", i);
    m->names[i] = strdup(t);
    if (!m->names[i]) {
      for (i = 0; i < m->nb_devices; i++) if (m->names[i]) free (m->names[i]);
      free(m->names);
      m->names = NULL;
      m->nb_devices = 0;
      return -1;
    }
  }

  return 0;
#endif /* WINDOZ */
}

void free_midi_devices(midi *m)
{
  int i;
  if (m->nb_devices) {
    for (i = 0; i < m->nb_devices; i++) free(m->names[i]);
    free(m->names);
  }
  m->nb_devices = 0;
  m->names = NULL;
}

void init_midi(midi *m)
{
  int i;

  memset(m, 0, sizeof(midi));

  /* all channels are off */
  for (i = 0; i < 16; i++) m->channels[i].remains = -1;

#ifdef ALSA
  /* note that we never close, maybe it is a problem? normally at program's
   * exit it should close, no?
   */
  if (snd_seq_open(&m->alsa_seq, "default", SND_SEQ_OPEN_OUTPUT, 0) < 0)
    goto err;
  snd_seq_set_client_name(m->alsa_seq, "Gcomposer");
  /* We create an output port on which we will WRITE, so logically
   * we have to open it with READ capability.
   * How clever! (ALSA sux)
   */
  m->alsa_port = snd_seq_create_simple_port(m->alsa_seq, "Gcomposer output",
                     SND_SEQ_PORT_CAP_READ | SND_SEQ_PORT_CAP_SUBS_READ,
                     SND_SEQ_PORT_TYPE_MIDI_GENERIC);
  if (m->alsa_port < 0) goto err;
  return;
err:
  /* nothing special to do, I just love so much gotos that I can't stop
   * puttting some everywhere!
   */
  /* oh yes maybe I should do a close for if create_simple_port fails */
  /* oh fuck, it does not matter */
  m->alsa_seq = NULL; /* so that later code can see ALSA failed */
  return;
#endif /* ALSA */
}

/* return -1 if write error, -2 if not all data written, 0 if ok */
int play_note_midi(midi *m, int note, int len, int vol)
{
  char p[3];
  int i;
  int ret;
  int imin;

  dprintf("midi play note %d len %d\n", note, len);
  /* try to get a free buffer */
  for (i = 0; i < 16; i++) if (m->channels[i].remains == -1) break;

  if (i != 16) goto found_channel;

  /* no free channel, heck, free the smallest */
  imin = 0;
  for (i = 1; i < 16; i++)
    if (m->channels[i].remains < m->channels[imin].remains)
      imin = i;

  i = imin;

  /* we have to stop the note */
  if (m->channels[i].note != -1) {
#ifndef WINDOZ
#ifdef ALSA
    if (!m->is_alsa) {
#endif
      p[0] = 0x80;
      p[1] = m->channels[i].note;
      p[2] = 0;

      if ((ret = write(m->fd, p, 3)) != 3) {
        if (ret >= 0) return -2;
        return ret;
      }
#ifdef ALSA
    } else {
      alsa_stop(m->alsa_seq, m->alsa_port, 0, m->channels[i].note, 0);
    }
#endif
#else /* WINDOZ */
    {
      unsigned long val;
      val = 0x80 | ((m->channels[i].note & 255) << 8);
      midiOutShortMsg(midi_windows, val);
    }
#endif /* WINDOZ */
    dprintf("midi stop ok %d write %2.2x %2.2x %2.2x\n",
            note, (unsigned)p[0], (unsigned)p[1], (unsigned)p[2]);
  }

found_channel:
  m->channels[i].remains = len;
  m->channels[i].note = note;

  if (note != -1) {
#ifndef WINDOZ
#ifdef ALSA
    if (!m->is_alsa) {
#endif
      p[0] = 0x90;
      p[1] = note;
      p[2] = vol;

      if ((ret = write(m->fd, p, 3)) != 3) {
        if (ret >= 0) return -2;
        return ret;
      }
#ifdef ALSA
    } else {
      alsa_play(m->alsa_seq, m->alsa_port, 0, note, vol);
    }
#endif
#else /* WINDOZ */
    {
      unsigned long val;
      val = 0x90 | ((note & 255) << 8) | (vol << 16);
      midiOutShortMsg(midi_windows, val);
    }
#endif /* WINDOZ */
    dprintf("midi play ok %d write %2.2x %2.2x %2.2x\n",
            note, (unsigned)p[0], (unsigned)p[1], (unsigned)p[2]);
  }

  return 0;
}

/* return -1 if write error, -2 if not all data written, 0 if ok */
int stop_finished_notes_midi(midi *m)
{
  char buf[16*3];
  int size = 0;
  int i;
  int ret;

  for (i = 0; i < 16; i++) {
    if (m->channels[i].remains != 0) continue;
    dprintf("midi stop note %d\n", m->channels[i].note);
    m->channels[i].remains = -1;
    if (m->channels[i].note != -1) {
      buf[size] = 0x80;
      buf[size+1] = m->channels[i].note;
      buf[size+2] = 0;
      size += 3;
#ifdef WINDOZ
      {
        unsigned long val;
        val = 0x80 | ((m->channels[i].note & 255) << 8);
        midiOutShortMsg(midi_windows, val);
      }
#else /* WINDOZ */
#ifdef ALSA
      if (m->is_alsa) {
        alsa_stop(m->alsa_seq, m->alsa_port, 0, m->channels[i].note, 0);
      }
#endif
#endif /* WINDOZ */
    }
  }

#ifndef WINDOZ
#ifdef ALSA
  if (!m->is_alsa)
#endif
    if (size)
      if ((ret = write(m->fd, buf, size)) != size) {
        if (ret >= 0) return -2;
        return ret;
      }
#endif /* WINDOZ */

#ifdef DEBUG_ON
  {
    int i;
    printf("midi stop size %d (", size);
    for (i=0; i<size; i++) printf("%2.2x ", (unsigned)buf[i]);
    printf(")\n");
    fflush(stdout);
  }
#endif

  return 0;
}

static void unset_midi_timer(midi *m)
{
  if (m->timer_id == 0) return;
  gtk_wrapper_remove_callback(m->timer_id);
  m->timer_id = 0;
}

static int midi_callback(midi *m)
{
  m->callback(m->data);
  return 0;
}

int set_midi_timer(midi *m)
{
  int i;
  int timeout;
  int set = 0;

  for (i = 0; i < 16; i++) {
    if (m->channels[i].remains <= 0) continue;
    if (!set) { set = 1; timeout = m->channels[i].remains; }
    else if (m->channels[i].remains < timeout)
      timeout = m->channels[i].remains;
  }

  if (!set) return 0;

  /* decrase time of channels */
  for (i = 0; i < 16; i++) {
    if (m->channels[i].remains <= 0) continue;
    m->channels[i].remains -= timeout;
  }

  m->shot_remain = timeout;

  m->timer_id = gtk_wrapper_timeout_add(timeout / 1000, midi_callback, m);

  return 0;
}

/* return -1 if error, -2 if not all data written, 0 if ok */
int start_midi(midi *m, int instrument)
{
  int i;
  char p[2];
  int ret;

#ifndef WINDOZ
  /* if alsa wanted but no alsa seq here then error */
#ifdef ALSA
  if (m->is_alsa && !m->alsa_seq) return -1;
#endif

#ifdef ALSA
  if (!m->is_alsa) {
#endif
    m->fd = open(m->device, O_WRONLY);
    if (m->fd < 0) return m->fd;
#ifdef ALSA
  }
#endif
#else /* WINDOZ */
  if (midiOutOpen(&midi_windows, midi_windows_device, 0, 0,
                  CALLBACK_NULL)) return -1;
#endif /* WINDOZ */

  /* all channels are off */
  for (i = 0; i < 16; i++) m->channels[i].remains = -1;

  /* select instrument */
  p[0] = 0xc0;
  p[1] = instrument;

#ifndef WINDOZ
#ifdef ALSA
  if (!m->is_alsa) {
#endif
    if ((ret = write(m->fd, p, 2)) != 2) {
      close(m->fd);
      if (ret >= 0) return -2;
      return ret;
    }
#ifdef ALSA
  } else {
    alsa_inst(m->alsa_seq, m->alsa_port, 0, instrument);
  }
#endif
#else /* WINDOZ */
  {
    unsigned long val;
    val = 0xc0 | ((instrument & 255) << 8);
    midiOutShortMsg(midi_windows, val);
  }
#endif /* WINDOZ */

  dprintf("midi selct inst ok write %2.2x %2.2x\n",
          (unsigned)p[0], (unsigned)p[1]);

  return 0;
}

void stop_midi(midi *m)
{
  char buf[16*3];
  int size = 0;
  int i;

  /* end all 'on notes' */
  for (i = 0; i < 16; i++) {
    if (m->channels[i].remains == -1) continue;
    m->channels[i].remains = -1;
    if (m->channels[i].note != -1) {
      buf[size] = 0x80;
      buf[size+1] = m->channels[i].note;
      buf[size+2] = 0;
      size += 3;
#ifdef WINDOZ
      {
        unsigned long val;
        val = 0x80 | ((m->channels[i].note & 255) << 8);
        midiOutShortMsg(midi_windows, val);
      }
#else /* WINDOZ */
#ifdef ALSA
      if (m->is_alsa) {
        alsa_stop(m->alsa_seq, m->alsa_port, 0, m->channels[i].note, 0);
      }
#endif
#endif /* WINDOZ */
    }
  }

  /* unchecked write, it may be a problem in rare occasions */
#ifndef WINDOZ
#ifdef ALSA
  if (!m->is_alsa)
#endif
    if (size) write(m->fd, buf, size);
#endif /* WINDOZ */

#ifdef DEBUG_ON
  {
    int i;
    printf("midi write size %d in stop_midi\n", size);
    for (i=0; i<size; i++) printf("%2.2x ", (unsigned)buf[i]);
    printf(")\n");
    fflush(stdout);
  }
#endif

#ifndef WINDOZ
#ifdef ALSA
  if (!m->is_alsa)
#endif
    close(m->fd);
#else /* WINDOZ */
  dprintf("MIDI closed!\n");
  midiOutClose(midi_windows);
#endif /* WINDOZ */

  unset_midi_timer(m);
}

/* force the usage of a specific device (/dev/SOMETHING generally) */
void force_midi_device(midi *m, char *name)
{
  m->device = name;
#ifdef WINDOZ
  /* what to do? */
  midi_windows_device = (unsigned long)-1;
#endif /* WINDOZ */
}

/* n > 99 || n < 0 is forbidden, the app has to ensure */
void set_midi_device(midi *m, int n)
{
  /* this static char lets the midi.c file be unsable in a multi-something
   * environment, but normally the program is not multi-something
   */
  static char name[12];
  sprintf(name, "/dev/midi%2.2d", n);
  m->device = name;
#ifdef WINDOZ
  midi_windows_device = n;
#endif /* WINDOZ */
}

void set_midi_callback(midi *m, void *data, void (*callback)(void *))
{
  m->data = data;
  m->callback = callback;
}
