/* gcomposer: a little tool to write music.
 * http://sed.free.fr/gcomposer
 * Released in the public domain (freedom has no license).
 */

#include "play.h"

#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "doc.h"
#include "gtk_wrapper.h"
#include "error.h"
#include "oss.h"
#include "draw.h"
#include "dim.h"
#include "size.h"
#include "dprintf.h"

/* middle A = 440 Hz
 * To get 1/2 tone higher, you have to do: cur_freq * 2^(1/12)
 * because you have 12 1/2 tones per octave and the freq
 * of higher A is freq of middle A * 2
 * You deduce the values of this array by applying the above rule
 * again and again.
 * See gen_freq.c for the code to produce this array.
 * This array only gets 1 octave. You then multiply/divide by multiples
 * of 2 to get the note you want.
 */
static float freqs[7][3] = {
#include "freqs.h"
};

/* same stuff but indicates midi pitch
 * 60 is for middle C
 */
static float pitchs[7][3] = {
#include "pitchs.h"
};

/* return -1 if error, allocated channel number if ok */
static int alloc_channel(play *p)
{
  if (p->size == p->maxsize) {
    play_note *new;
    int ns = p->maxsize + 32;
    if (p->maxsize)
      new = realloc(p->data, ns * sizeof(play_note));
    else
      new = malloc(ns * sizeof(play_note));
    if (!new) {
      set_error(OOM);
      return -1;
    }
    p->data = new;
    p->maxsize = ns;
  }
  p->size++;
  return p->size-1;
}

/* 1 ok 0 error */
/* give setup_chord == -1 to play
 * give a given chord to simulate play up to this chord (so that # and b
 * stuff are later on correctly handled)
 */
static int start_chord(doc *d, int setup_chord)
{
  play *p = &d->p;
  chord *c;
  int i;
  int channel_number;
  play_note *channel;
  int global_remains_up = -1;
  int global_remains_down = -1;
  int remains;
  elem *e;
  float t;
  int note_chord_processed = 0;
  int midi_pitch;

  /* how this function works:
   * we process all MESURE_BREAK, TONALITE and CLEF stuff
   * then one NOTES_CHORD
   * then all MESURE_BREAK, TONALITE and CLEF stuff up to the next NOTES_CHORD
   * so that in the visual feedback only NOTES_CHORD are highlighted
   * this explains the note_chord_processed variable and the
   * go_to_next_note_chords label (we could use better the while
   * and switch statements, but the heck, gotos are nice)
   *
   * when setup_chord is passed with a value != -1 we are in a "simulation"
   * mode which is used to start the playing at any point in a line
   * we need to simulate previous # and b settings that may impact
   * the wanted note
   */

go_to_next_notes_chord:
  /* process MESURE_BREAK, TONALITE and CLEF */
  while (1) {
    if (p->line >= d->size) {
      /* if we didn't process any chord, save this fact */
      if (!note_chord_processed)
        if (setup_chord == -1) {
          p->processed_chord = -1;
          p->processed_line = -1;
        }
      return 1;
    }
    if (setup_chord != -1) if (p->chord >= setup_chord) return 1;
    if (p->chord >= d->data[p->line].u.music.size) goto go_next_line;
    c = &CHORD(d, p->line, p->chord);
    switch (c->type) {
      case NOTES_CHORD:
        if (note_chord_processed) return 1;
        goto notes_chord;
      case MESURE_BREAK:
        for (i = 0; i < KNOWN_NOTES; i++)
          p->current_mode[i] = p->global_mode[i];
        break;
      case TONALITE: {
        int pos;
        int alteration;
        switch (c->u.tonalite.alteration) {
          case NORMAL: /* we can't come here */ alteration = 0; break;
          case DIESE: alteration = 1; break;
          case BEMOL: alteration = -1; break;
          case BECARRE: alteration = 0; break;
        }
        pos = c->u.tonalite.position + p->delta_clef;
        while (pos <= MAX_KNOWN_NOTE) {
          p->global_mode[-MIN_KNOWN_NOTE + pos] = alteration;
          p->current_mode[-MIN_KNOWN_NOTE + pos] = alteration;
          pos += 7;
        }
        pos = c->u.tonalite.position - 7 + p->delta_clef;
        while (pos >= MIN_KNOWN_NOTE) {
          p->global_mode[-MIN_KNOWN_NOTE + pos] = alteration;
          p->current_mode[-MIN_KNOWN_NOTE + pos] = alteration;
          pos -= 7;
        }
        break;
      }
      case CLEF: {
        /* reset everything */
        for (i = 0; i < KNOWN_NOTES; i++)
          p->current_mode[i] = p->global_mode[i] = 0;
        p->delta_clef = -c->u.clef.position;
        switch (c->u.clef.type) {
          /*
           * Hope these values are correct:
           *               +- middle C = -4
           *               |                     +- G 8vb = 7
           *               |                     |
           * C D E F G A B C D E F G A B C D E F G
           *       | |             |
           *       | |             +- standard G = 0
           *       | +- G 8va = -7
           *       +- F below = -8
           */
          case CLEF_G: p->delta_clef += 0; break;
          case CLEF_F: p->delta_clef += -8; break;
          case CLEF_C: p->delta_clef += -4; break;
          case CLEF_G_8va: p->delta_clef += -7; break;
          case CLEF_G_8vb: p->delta_clef += 7; break;
        }
        break;
      }
    }
    p->chord++;
    if (p->chord >= d->data[p->line].u.music.size) {
go_next_line:
      if (setup_chord != -1) return 1;
      /* we zap non-music lines */
      do {
        p->line++;
      } while (p->line < d->size && d->data[p->line].type != LINE_MUSIC);
      p->chord = 0;
      for (i=0; i < KNOWN_NOTES; i++) {
        p->global_mode[i] = 0;
        p->current_mode[i] = 0;
      }
      p->delta_clef = 0;
    }
  } /* while (1) */

notes_chord: /* I can't stop using gotos */
  for (i = 0; i < c->u.chord.size; i++) {
    e = &c->u.chord.data[i];
    dprintf("note %d\n", e->u.note.pitch);
    /* calculate the duration */
    /* tempo * NOIRE = 44100 * 60
     * x = NOIRE * t
     * tempo * (x/t) = 44100 * 60
     * x = 44100 * 60 * t / tempo
     */
    switch (e->type) {
      case NOTE:
        switch (e->u.note.duration) {
          case RONDE: t = 4.; break;
          case BLANCHE: t = 2.; break;
          case NOIRE: t = 1.; break;
          case CROCHE: t = .5; break;
          case DCROCHE: t = .25; break;
          case DDCROCHE: t = .125; break;
          case DDDCROCHE: t = .0625; break;
          case DDDDCROCHE: t = .03125; break;
        }
        if (e->u.note.nb_dots == 3) t = t + t/2. + t/4. + t/8.;
        else if (e->u.note.nb_dots == 2) t = t + t/2. + t/4.;
        else if (e->u.note.nb_dots == 1) t = t + t/2.;
        break;
      case SILENCE:
        switch (e->u.silence.duration) {
          case RONDE: t = 4.; break;
          case BLANCHE: t = 2.; break;
          case NOIRE: t = 1.; break;
          case CROCHE: t = .5; break;
          case DCROCHE: t = .25; break;
          case DDCROCHE: t = .125; break;
          case DDDCROCHE: t = .0625; break;
          case DDDDCROCHE: t = .03125; break;
        }
        if (e->u.silence.nb_dots == 3) t = t + t/2. + t/4. + t/8.;
        else if (e->u.silence.nb_dots == 2) t = t + t/2. + t/4.;
        else if (e->u.silence.nb_dots == 1) t = t + t/2.;
        break;
    }
    if (d->options.ptype == PLAY_OSS)
      remains = 44100. * 60. * t / p->tempo;
    else
      remains = 1000000. * 60. * t / p->tempo;
#if 0
    if (i == 0) global_remains = remains;
    if (remains < global_remains) global_remains = remains;
#endif
    /* this if stuff should be better done */
    if (e->type == SILENCE) {
      if (global_remains_up == -1 || remains < global_remains_up)
        global_remains_up = remains;
      if (global_remains_down == -1 || remains < global_remains_down)
        global_remains_down = remains;
    } else {
      if (e->u.note.stem_down) {
        if (global_remains_down == -1 || remains < global_remains_down)
          global_remains_down = remains;
      } else {
        if (global_remains_up == -1 || remains < global_remains_up)
          global_remains_up = remains;
      }
    }
    /* allocate a channel if there is a note (including silence) to play */
    /* allocate if we are not in setup mode */
    if (setup_chord == -1 && d->options.ptype == PLAY_OSS) {
      if ((channel_number = alloc_channel(p)) == -1) return 0;
      channel = &p->data[channel_number];
      channel->total = remains;
      channel->current = 0;
    }
    switch (e->type) {
      case NOTE: {
        int pitch = e->u.note.pitch + p->delta_clef;
        /* a bit brutal but it should not happen */
        if (pitch < MIN_KNOWN_NOTE) pitch = MIN_KNOWN_NOTE;
        if (pitch > MAX_KNOWN_NOTE) pitch = MAX_KNOWN_NOTE;
        /* update current alteration stuff */
        /* this may not be what musicians expect. Heck, I'll fix this stuff
         * later on if there is a problem. (The problem is that if you
         * have a diese at beginning of a line and if you add another
         * diese later on the line, what note is it? Suppose you have a
         * diese for G, later on you have G#, what to play? G# or F?
         * And if you have Gb, what to play? Gb or G? Here we simply
         * discard the symbol at beginning of line, so we would play G#
         * and Gb.)
         * Note also that only the G of the current line is updated. If later
         * on the same mesure you have a lower G, it will keep its
         * alteration.
         * Some engravers will suppose that it changes every G. Both
         * possibilities exist in real-life, so I don't know...
         */
        switch (e->u.note.alteration) {
          case NORMAL: break;
          case DIESE: p->current_mode[-MIN_KNOWN_NOTE + pitch] = 1; break;
          case BEMOL: p->current_mode[-MIN_KNOWN_NOTE + pitch] = -1; break;
          case BECARRE: p->current_mode[-MIN_KNOWN_NOTE + pitch] = 0; break;
        }
        /* get frequency if not in setup mode */
        if (setup_chord == -1) {
          if (d->options.ptype == PLAY_OSS) {
            int pt = pitch;
            switch (p->current_mode[-MIN_KNOWN_NOTE + pitch]) {
              case -1:
                channel->freq = freqs[(-MIN_KNOWN_NOTE*7 + pitch) % 7][0];
                break;
              case 0:
                channel->freq = freqs[(-MIN_KNOWN_NOTE*7 + pitch) % 7][1];
                break;
              case 1:
                channel->freq = freqs[(-MIN_KNOWN_NOTE*7 + pitch) % 7][2];
                break;
            }
            /* ugly, slow, may change in the future */
            while (pt > 6) {
              pt -= 7;
              channel->freq *= 2;
            }
            while (pt < 0) {
              pt += 7;
              channel->freq /= 2;
            }
          }
          /* play the note if midi mode */
          if (d->options.ptype == PLAY_MIDI ||
              d->options.ptype == PLAY_MIDI_ALSA) {
            int pitch = e->u.note.pitch + p->delta_clef;
            int pt;
            /* a bit brutal but it should not happen */
            if (pitch < MIN_KNOWN_NOTE) pitch = MIN_KNOWN_NOTE;
            if (pitch > MAX_KNOWN_NOTE) pitch = MAX_KNOWN_NOTE;
            switch (p->current_mode[-MIN_KNOWN_NOTE + pitch]) {
              case -1:
                midi_pitch = pitchs[(-MIN_KNOWN_NOTE*7 + pitch) % 7][0];
                break;
              case 0:
                midi_pitch = pitchs[(-MIN_KNOWN_NOTE*7 + pitch) % 7][1];
                break;
              case 1:
                midi_pitch = pitchs[(-MIN_KNOWN_NOTE*7 + pitch) % 7][2];
                break;
            }
            pt = pitch;
            /* ugly, slow, may change in the future */
            while (pt > 6) {
              pt -= 7;
              midi_pitch += 12;
            }
            while (pt < 0) {
              pt += 7;
              midi_pitch -= 12;
            }
            midi_pitch += d->options.midi_adjust;
            if (midi_pitch < 0) midi_pitch = 0;
            if (midi_pitch > 127) midi_pitch = 127;
            if (play_note_midi(&d->midi, midi_pitch, remains,
                               d->options.midi_volume) < 0) {
#ifdef ALSA
              if (d->midi.is_alsa)
                set_error(MIDI_ERROR, "ALSA MIDI device");
              else
#endif
                set_error(MIDI_ERROR, d->midi.device);
              return 0;
            }
          } /* if (d->options.ptype == PLAY_MIDI || PLAY_MIDI_ALSA) */
        } /* if (setup_chord == -1) */
        break;
      }
      case SILENCE:
        if (setup_chord == -1) {
          if (d->options.ptype == PLAY_OSS)
            channel->freq = 0;
          else if (d->options.ptype == PLAY_MIDI ||
                   d->options.ptype == PLAY_MIDI_ALSA)
            if (play_note_midi(&d->midi, -1, remains,
                               d->options.midi_volume) < 0) {
#ifdef ALSA
              if (d->midi.is_alsa)
                set_error(MIDI_ERROR, "ALSA MIDI device");
              else
#endif
                set_error(MIDI_ERROR, d->midi.device);
              return 0;
            }
        }
        break;
    } /* switch(e->type) */
  } /* for */

  if (setup_chord == -1) {
    p->processed_chord = p->chord;
    p->processed_line = p->line;
  }

  /* one of them may be unset */
  if (global_remains_up == -1) global_remains_up = global_remains_down;
  if (global_remains_down == -1) global_remains_down = global_remains_up;

  /* voice stuff, check the next chord (the next NOTES_CHORD which might
   * be very far away).
   * If it has only one direction, take this one.
   * If not, take min of both.
   * Rests don't count.
   * If no next chord, take also min.
   */
  {
    int next_line = p->line;
    int next_chord = p->chord + 1;

    /* find the next chord (if there is one)
     * if nothing found then next_chord == -1 after the loop
     */
    while (1) {
      /* end of doc and nothing found? */
      if (next_line >= d->size) {
        next_chord = -1;
        goto after_search_chord;
      }

      /* check the line */
      while (next_chord < d->data[next_line].u.music.size) {
        if (CHORD_TYPE(d, next_line, next_chord) == NOTES_CHORD)
          goto after_search_chord;
        next_chord++;
      }

      /* nothing found, go to next line */
      next_chord = 0;

      /* zap non music lines */
      do {
        next_line++;
      } while (next_line < d->size && d->data[next_line].type != LINE_MUSIC);
    }
after_search_chord:

    /* if no next chord take the min of both remains */
    /* (we could take the maximum of all notes because it's the last
     * chord, this would remove a KNOWN_BUG, but hey...)
     */
    if (next_chord == -1) {
      if (global_remains_down < global_remains_up)
        global_remains_up = global_remains_down;
    } else {
      /* it's a NOTES_CHORD check the voices it has */
      int has_up = 0;
      int has_down = 0;
      int l;

      for (l = 0; l < CHORD_SIZE(d, next_line, next_chord); l++) {
        if (NOTE_TYPE(d, next_line, next_chord, l) != NOTE) continue;
        if (NOTE_STEM_DOWN(d, next_line, next_chord, l))
          has_down = 1;
        else
          has_up = 1;
      }

      /* if no up and no down do as if it has both */
      if (!has_up && !has_down) {
        has_up = 1;
        has_down = 1;
      }

      /* pickup min of what is valid */
      if (!has_up) {
        global_remains_up = global_remains_down;
      } else {
        if (has_down) {
          if (global_remains_down < global_remains_up)
            global_remains_up = global_remains_down;
        }
      }
    }
  }

  p->remains = global_remains_up;

  /* we are done with this chord */
  p->chord++;
  if (p->chord >= d->data[p->line].u.music.size) {
    if (setup_chord != -1) return 1;
    /* we zap non-music lines */
    do {
      p->line++;
    } while (p->line < d->size && d->data[p->line].type != LINE_MUSIC);
    p->chord = 0;
    for (i=0; i < KNOWN_NOTES; i++) {
      p->global_mode[i] = 0;
      p->current_mode[i] = 0;
    }
    p->delta_clef = 0;
  }
  note_chord_processed = 1;
  goto go_to_next_notes_chord;
}

static short buffer[2048];
static float bfloat[2048];

static void put(int pos, int size, play_note *note)
{
  int i;
  float *p = bfloat + pos;
  dprintf("put pos %d size %d total %d current %d freq %f\n",
          pos, size, note->total, note->current, note->freq);
  if (note->freq == 0) return;
  for (i = 0; i < size; i++, p++)
    *p += 10000. * cos(2. * M_PI * (i + note->current) * note->freq/44100.);
}

static void play_a_bit(doc *d)
{
  int total = 2048;
  int max;
  int i;
  play *p = &d->p;
  int previous_line;
  int previous_chord;
  int current_line;
  int current_chord;

#if 0
  /* barbarian invalidation - rewrite */
  if (d->p.starting_chord[d->p.current_chord].line != -1)
    invalidate_line(d, d->p.starting_chord[d->p.current_chord].line);
#endif
  previous_line = d->p.starting_chord[d->p.current_chord].line;
  previous_chord = d->p.starting_chord[d->p.current_chord].chord;

  d->p.starting_chord[d->p.current_chord].line = d->p.processed_line;
  d->p.starting_chord[d->p.current_chord].chord = d->p.processed_chord;

  d->p.current_chord++;
  if (d->p.current_chord == NB_BUFFERS) d->p.current_chord = 0;

  current_line = d->p.starting_chord[d->p.current_chord].line;
  current_chord = d->p.starting_chord[d->p.current_chord].chord;

  if (current_line != previous_line || current_chord != previous_chord) {
    if (previous_line != -1) {
//      invalidate_line(d, previous_line);
      chord *c = &d->data[previous_line].u.music.data[previous_chord];
      ensure_line_positioned(d, previous_line);
      draw_line(d, previous_line,
                c->dim.x + LEFT_MARGIN,
                c->dim.x + LEFT_MARGIN + c->dim.width - 1);
    }
    if (current_line != -1) {
//      invalidate_line(d, current_line);
      chord *c = &d->data[current_line].u.music.data[current_chord];
      show_chord(d, current_line, current_chord);
//      ensure_line_positioned(d, current_line);
      draw_line(d, current_line,
                c->dim.x + LEFT_MARGIN,
                c->dim.x + LEFT_MARGIN + c->dim.width - 1);
    }
  }

#if 0
  if (d->p.starting_chord[d->p.current_chord].line != -1) {
    show_chord(d, d->p.starting_chord[d->p.current_chord].line,
               d->p.starting_chord[d->p.current_chord].chord);
    invalidate_line(d, d->p.starting_chord[d->p.current_chord].line);
  }
#endif

  memset(bfloat, 0, 2048 * sizeof(float));

  while (total) {
    max = total;
    /* p->remains might be 0 if there is no more chord to play */
    if (p->remains && max > p->remains) max = p->remains;
    /* calculate minimum max */
    for (i = 0; i < p->size; i++) {
      if (max >= p->data[i].total - p->data[i].current)
        max = p->data[i].total - p->data[i].current;
    }
    /* put max samples and free empty channels */
    for (i = 0; i < p->size; i++) {
      put(2048-total, max, &p->data[i]);
      p->data[i].current += max;
      if (p->data[i].current == p->data[i].total) {
        /* slow */
        memmove(p->data+i, p->data+i+1, sizeof(play_note) * (p->size-i-1));
        i--;
        p->size--;
      }
    }
    /* we are done */
    total -= max;
    if (p->remains) p->remains -= max;
    if (p->remains == 0) {
      if (!start_chord(d, -1)) {
        stop_play(d);
        return;
      }
    }
  }
  for (i = 0; i < 2048; i++)
    if (bfloat[i] < -32767)
      buffer[i] = -32767;
    else if (bfloat[i] > 32767)
      buffer[i] = 32767;
    else
      buffer[i] = bfloat[i];

#ifndef WINDOZ
  if (!play_oss(buffer, 4096)) stop_play(d);
#endif
}

static void reset_global_mode(doc *d)
{
  int i;

  /* reset global_mode and current_mode */
  for (i=0; i < KNOWN_NOTES; i++) {
    d->p.global_mode[i] = 0;
    d->p.current_mode[i] = 0;
  }
  d->p.delta_clef = 0;
}

static void midi_callback(doc *d)
{
  int chord_changed;
  chord *c;
  int prev_line = d->p.processed_line;
  int prev_chord = d->p.processed_chord;
  if (stop_finished_notes_midi(&d->midi) < 0) {
    stop_play(d);
    return;
  }
  d->p.remains -= d->midi.shot_remain;
  chord_changed = d->p.remains == 0;
  if (chord_changed)
    if (!start_chord(d, -1)) {
      stop_play(d);
      return;
    }
  if (set_midi_timer(&d->midi) < 0) {
    stop_play(d);
#ifdef ALSA
    if (d->midi.is_alsa)
      set_error(MIDI_ERROR, "ALSA MIDI device");
    else
#endif
      set_error(MIDI_ERROR, d->midi.device);
    return;
  }
  if (chord_changed) {
    /* redraw previous chord */
    if (prev_line != -1) {
      c = &d->data[prev_line].u.music.data[prev_chord];
      ensure_line_positioned(d, prev_line);
      draw_line(d, prev_line,
                c->dim.x + LEFT_MARGIN,
                c->dim.x + LEFT_MARGIN + c->dim.width - 1);
    }
    /* redraw current chord and show it */
    if (d->p.processed_line != -1) {
      c = &d->data[d->p.processed_line].u.music.data[d->p.processed_chord];
      show_chord(d, d->p.processed_line, d->p.processed_chord);
      draw_line(d, d->p.processed_line,
                c->dim.x + LEFT_MARGIN,
                c->dim.x + LEFT_MARGIN + c->dim.width - 1);
    }
  }
}

int start_play(doc *d, int start_line, int starting_chord, int tempo)
{
  int i;

#ifdef WINDOZ
  /* we only want midi */
  if (d->options.ptype == PLAY_OSS) {
    set_error(OSS_ERROR);
    return 0;
  }
#endif /* WINDOZ */

  if (d->options.ptype == PLAY_OSS) {
    /* 12 is for 2^12 = 4096: the size of one audio buffer */
    if (!init_oss(d->options.audio_oss_file, 44100, 16, 0,
                  NB_BUFFERS - 1, 12, PLAY_SOUND))
      return 0;

    d->p.input_id = gtk_wrapper_add_audio_callback(get_oss_sound_fd(),
                                                   play_a_bit, d);

    /* reset the memorized chords used for display synchronization */
    d->p.current_chord = 0;
    for (i = 0; i < NB_BUFFERS; i++)
      d->p.starting_chord[i].line = -1;
  }

  d->p.tempo = tempo;
  d->p.line = start_line;
  d->p.chord = 0;

  /* ensure that the current # and b settings are correct */
  reset_global_mode(d);
  while (d->p.chord < starting_chord) start_chord(d, starting_chord);

  d->p.chord = starting_chord;
  d->p.remains = 0;

  if (d->options.ptype == PLAY_MIDI ||
      d->options.ptype == PLAY_MIDI_ALSA) {
    if (d->options.midi_device == -1)
      force_midi_device(&d->midi, d->options.audio_midi_file);
    else
      set_midi_device(&d->midi, d->options.midi_device);
#ifdef ALSA
    d->midi.is_alsa = d->options.ptype == PLAY_MIDI_ALSA;
#endif
    if (start_midi(&d->midi, d->options.midi_instrument) < 0) {
#ifdef ALSA
      if (d->midi.is_alsa)
        set_error(MIDI_ERROR, "ALSA MIDI device");
      else
#endif
        set_error(MIDI_ERROR, d->midi.device);
      return 0;
    }
    set_midi_callback(&d->midi, d, (void(*)(void*))midi_callback);
  }

  d->p.processed_line = -1;
  if (!start_chord(d, -1)) {
    stop_play(d);
    return 0;
  }

#ifndef WINDOZ
  if (d->options.ptype == PLAY_OSS) do play_a_bit(d); while (!oss_full());
#endif /* WINDOZ */

  if (d->options.ptype == PLAY_MIDI || d->options.ptype == PLAY_MIDI_ALSA) {
    /* bad reuse of 'current_chord' to know if we are in play mode or not
     * when calling in_play
     */
    d->p.current_chord = 0;
    if (set_midi_timer(&d->midi) < 0) {
      stop_play(d);
#ifdef ALSA
      if (d->midi.is_alsa)
        set_error(MIDI_ERROR, "ALSA MIDI device");
      else
#endif
        set_error(MIDI_ERROR, d->midi.device);
      return 0;
    }
    /* redraw current chord and show it */
    if (d->p.processed_line != -1) {
      chord *c = &d->data[d->p.processed_line].u.music.
                     data[d->p.processed_chord];
      show_chord(d, d->p.processed_line, d->p.processed_chord);
      draw_line(d, d->p.processed_line,
                c->dim.x + LEFT_MARGIN,
                c->dim.x + LEFT_MARGIN + c->dim.width - 1);
    }
  }

  return 1;
}

int stop_play(doc *d)
{
  int current_line;
  int current_chord;
  play *p = &d->p;

  if (d->options.ptype == PLAY_OSS) {
    if (p->maxsize) free(p->data);
    p->maxsize = 0;
    p->size = 0;
    p->free = -1;
    p->line = -1;
    /* we redraw the current chord */
    current_line = d->p.starting_chord[d->p.current_chord].line;
    current_chord = d->p.starting_chord[d->p.current_chord].chord;
    /* before we redraw, tell the drawer we are not in play mode anymore */
    p->current_chord = -1;
    /* now let's redraw (if we have to) */
    if (current_line != -1) {
      chord *c = &d->data[current_line].u.music.data[current_chord];
      draw_line(d, current_line,
                c->dim.x + LEFT_MARGIN,
                c->dim.x + LEFT_MARGIN + c->dim.width - 1);
    }
  //  invalidate_line(d, d->p.starting_chord[d->p.current_chord].line);
    gtk_wrapper_remove_callback(p->input_id);
#ifndef WINDOZ
    close_oss();
#endif
  }
  if (d->options.ptype == PLAY_MIDI || d->options.ptype == PLAY_MIDI_ALSA) {
    p->current_chord = -1;
    if (p->processed_line != -1) {
      chord *c = &d->data[p->processed_line].u.music.data[p->processed_chord];
      draw_line(d, p->processed_line,
                c->dim.x + LEFT_MARGIN,
                c->dim.x + LEFT_MARGIN + c->dim.width - 1);
    }
    stop_midi(&d->midi);
  }

  return 1;
}

int in_play(doc *d, int line, int chord)
{
  if (d->options.ptype == PLAY_OSS) {
    return d->p.current_chord != -1 &&
           line == d->p.starting_chord[d->p.current_chord].line &&
           chord == d->p.starting_chord[d->p.current_chord].chord;
  }
  return d->p.current_chord != -1 &&
         d->p.processed_line == line &&
         d->p.processed_chord == chord;
}
