/* gcomposer: a little tool to write music.
 * http://sed.free.fr/gcomposer
 * Released in the public domain (freedom has no license).
 */

#include "thread.h"

#include <stdlib.h>

#include "error.h"
#include "states.h"

/* hell, "__thread" is used by gcc, so we must use "__thrd" */
typedef struct {
  struct _m_state **stack;
  int size;
  int maxsize;
} __thrd;

typedef struct {
  union {
    int free_next;
    __thrd t;
  } u;
} _thread;

typedef struct {
  _thread *data;
  int free;
  int size;
  int maxsize;
} thread_manager;

static thread_manager threads = {
  .free = -1,
  .maxsize = 0
};

/* -1 error, 0 ok */
int thread_push(thread t, struct _m_state *s)
{
  __thrd *cur = &threads.data[t].u.t;

  if (cur->size == cur->maxsize) {
    int new_size = cur->maxsize + 64;
    struct _m_state **new;
    if (cur->maxsize)
      new = realloc(cur->stack, sizeof(struct _m_state *) * new_size);
    else
      new = malloc(sizeof(struct _m_state *) * new_size);
    if (!new) { set_error(OOM); return -1; }
    cur->stack = new;
    cur->maxsize = new_size;
  }

  s->thread = t;
  cur->stack[cur->size] = s;
  cur->size++;

  return 0;
}

/* return NULL if nothing on top */
struct _m_state *thread_pop(thread t)
{
  __thrd *cur = &threads.data[t].u.t;
  if (cur->size == 0) return NULL;
  cur->size--;
  return cur->stack[cur->size];
}

/* return NULL if nothing on top */
struct _m_state *thread_top(thread t)
{
  __thrd *cur = &threads.data[t].u.t;
  if (cur->size == 0) return NULL;
  return cur->stack[cur->size-1];
}

thread new_thread(void)
{
  thread t;
  if (threads.free == -1) {
    int i;
    int newsize = threads.maxsize + 16;
    _thread *t;
    if (threads.maxsize)
      t = realloc(threads.data, newsize * sizeof(_thread));
    else
      t = malloc(newsize * sizeof(_thread));
    if (!t) { set_error(OOM); return -1; }
    threads.data = t;
    threads.free = threads.maxsize;
    threads.maxsize = newsize;
    for (i=threads.free; i < newsize - 1; i++)
      threads.data[i].u.free_next = i+1;
    threads.data[i].u.free_next = -1;
  }
  t = threads.free;
  threads.free = threads.data[t].u.free_next;
  threads.size++;

  threads.data[t].u.t.maxsize = 0;
  threads.data[t].u.t.size = 0;
  threads.data[t].u.t.stack = NULL;

  return t;
}

void thread_delete(thread t)
{
  __thrd *cur = &threads.data[t].u.t;
  int old_free = threads.free;

  if (cur->maxsize) free(cur->stack);
  cur->size = 0;
  cur->maxsize = 0;
  cur->stack = NULL;

  threads.free = t;
  threads.data[t].u.free_next = old_free;
  threads.size--;
}

void thread_close(void)
{
  int i;

  /* first of all, destroy the free list so we won't free unallocated stuff */
  i = threads.free;
  while (i != -1) {
    __thrd *cur = &threads.data[i].u.t;
    i = threads.data[i].u.free_next;
    cur->maxsize = 0;
  }

  /* now we can destroy stuff */
  for (i = 0; i < threads.maxsize; i++) {
    __thrd *cur = &threads.data[i].u.t;
    if (!cur->maxsize) continue;
    free(cur->stack);
    cur->size = 0;
    cur->maxsize = 0;
    cur->stack = NULL;
  }

  /* and finally remove the list itself */
  if (!threads.maxsize) return;

  free(threads.data);
  threads.free = -1;
  threads.maxsize = 0;
  threads.size = 0;
  threads.data = NULL;
}

int thread_top_type(thread t)
{
  __thrd *cur = &threads.data[t].u.t;
  return cur->stack[cur->size-1]->type;
}
